"""
Create embeddings for text

This module provides the `Embedding` class, which is used to generate embeddings for text using pre-trained models. 
The `Embedding` class supports different models such as MiniLM-L6-v2 and hkunlp/instructor-xl. The embeddings are 
generated for a specified text column in a DataFrame.

"""

from __future__ import annotations
import typing as t
import numpy as np

from loguru import logger
import polars as pl
import json
import requests

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import lazy_load_dep


@register_op
class Embedding(ColumnOp):
    """
    Column operation that generates embeddings for text using pre-trained models.

    Attributes:
        model (Literal["MiniLM-L6-v2", "instructor-xl", "mpnet-base-v2", "bge-large-zh-v1.5"]): The name of the pre-trained model to use.
        col_in_text (str): The name of the text column in the DataFrame.
        col_out (str): The name of the output column in the DataFrame.

    Raises:
        Exception: If the specified model is not supported.

    Returns:
        TYPE_TABLE_OUTPUT: A dictionary containing the generated embeddings.

    Example:
        ```
        import polars as pl
        from uptrain.operators import Embedding

        # Create a DataFrame
        df = pl.DataFrame({
            "text": ["This is the first sentence.", "Here is another sentence."]
        })

        # Create an instance of the Embedding class
        embedding_op = Embedding(model="MiniLM-L6-v2", col_in_text="text", col_out="embedding")

        # Set up the Embedding operator
        embedding_op.setup()

        # Generate embeddings for the text column
        embeddings = embedding_op.run(df)["output"]

        # Print the embeddings
        print(embeddings)
        ```

    Output:
        ```
        shape: (2,)
        Series: '_col_0' [list[f32]]
        [
                [0.098575, 0.056978, … -0.071038]
                [0.072772, 0.073564, … -0.043947]
        ]
        ```

    """

    model: t.Literal["MiniLM-L6-v2", "instructor-xl", "mpnet-base-v2", "bge-large-zh-v1.5", "instructor-large"]
    col_in_text: str = "text"
    col_out: str = "embedding"
    batch_size: int = 128

    def setup(self, settings: Settings):
        self._compute_method = settings.embedding_compute_method
        if settings.embedding_compute_method == 'local':
            if self.model == "instructor-xl":
                InstructorEmbedding = lazy_load_dep("InstructorEmbedding", "InstructorEmbedding")
                self._model_obj = InstructorEmbedding.INSTRUCTOR(self.model)  # type: ignore
            elif self.model == "MiniLM-L6-v2":
                sentence_transformers = lazy_load_dep("sentence_transformers", "sentence-transformers")
                self._model_obj = sentence_transformers.SentenceTransformer(
                    "sentence-transformers/all-MiniLM-L6-v2"
                )  # type: ignore
            elif self.model == "mpnet-base-v2":
                sentence_transformers = lazy_load_dep("sentence_transformers", "sentence-transformers")
                self._model_obj = sentence_transformers.SentenceTransformer(
                    "sentence-transformers/all-mpnet-base-v2"
                )
            elif self.model == "bge-large-zh-v1.5":
                sentence_transformers = lazy_load_dep("sentence_transformers", "sentence-transformers")
                self._model_obj = sentence_transformers.SentenceTransformer(
                    "BAAI/bge-large-zh-v1.5"
                )
            else:
                raise Exception(f"Embeddings model: {self.model} is not supported yet.")
        elif settings.embedding_compute_method == 'replicate':
            replicate = lazy_load_dep("replicate", "replicate")
            self._model_obj = replicate.Client(api_token=settings.replicate_api_token)
            if self.model == "mpnet-base-v2":
                self._model_url = "replicate/all-mpnet-base-v2:b6b7585c9640cd7a9572c6e129c9549d79c9c31f0d3fdce7baac7c67ca38f305"
            elif self.model == "bge-large-zh-v1.5":
                self._model_url = "nateraw/bge-large-en-v1.5:9cf9f015a9cb9c61d1a2610659cdac4a4ca222f2d3707a68517b18c198a9add1"
            else:
                raise Exception(f"Embeddings model: {self.model} is not supported yet.")
        elif settings.embedding_compute_method == 'api':
            self._model_obj = {
                'embedding_model_url': settings.embedding_model_url,
                'model': self.model,
                'authorization_key':settings.embedding_model_api_token
            }
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        text = data.get_column(self.col_in_text)
        if self.model in ["instructor-xl", "instructor-large", "bge-large-zh-v1.5"]:
            inputs = [
                ["Represent the sentence: ", x] for x in text
            ]
        elif self.model == "MiniLM-L6-v2" or self.model == "mpnet-base-v2":
            inputs = list(text)
        else:
            raise Exception("Embeddings model not supported")

        results = []
        BATCH_SIZE = self.batch_size
        if self.model == "bge-large-zh-v1.5":
            emb_length = 1024
        elif self.model in ["instructor-xl", "instructor-large"]:
            emb_length = 768
        for idx in range(int(np.ceil(len(inputs)/BATCH_SIZE))):
            if self._compute_method == "local":
                run_res = self._model_obj.encode(inputs[idx*BATCH_SIZE:(idx+1)*BATCH_SIZE])
            elif self._compute_method == "replicate":
                run_res = [x['embedding'] for x in self._model_obj.run(
                        self._model_url,
                        input = {"text_batch": json.dumps(inputs[idx*BATCH_SIZE:(idx+1)*BATCH_SIZE])}
                    )]
            elif self._compute_method == "api":
                try:
                    run_res = [x['embedding'] for x in requests.post(
                        self._model_obj['embedding_model_url'],
                        json={
                            'model': self.model,
                            'input': inputs[idx*BATCH_SIZE:(idx+1)*BATCH_SIZE]
                        },
                        headers={
                            'Authorization': f"Bearer {self._model_obj['authorization_key']}"
                        }
                    ).json()['data']]
                    emb_length = len(run_res[0])
                except:
                    run_res = []
                    for elem_idx in range(idx*BATCH_SIZE, (idx+1)*BATCH_SIZE):
                        if elem_idx < len(inputs):
                            try:
                                run_res.extend([x['embedding'] for x in requests.post(
                                    self._model_obj['embedding_model_url'],
                                    json={
                                        'model': self.model,
                                        'input': [inputs[elem_idx]]
                                    },
                                    headers={
                                        'Authorization': f"Bearer {self._model_obj['authorization_key']}"
                                    }
                                ).json()['data']])
                            except:
                                if len(run_res)!=0:
                                    run_res.append(run_res[-1])
                                else:
                                    run_res.append([0]*emb_length)
            results.extend(run_res)
            logger.info(f"Running batch: {idx} out of {int(np.ceil(len(inputs)/BATCH_SIZE))} for operator Embedding")
        return {"output": data.with_columns([pl.Series(results).alias(self.col_out)])}



import numpy as np

from typing import Any

from uptrain.v0.core.classes.measurables import Measurable


class RecHitRateMeasurable(Measurable):
    """Class that computes the recommendation hit-rate of a model's predictions."""

    def __init__(self, framework) -> None:
        super().__init__(framework)

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> Any:
        acc = [1 if gt in outputs[i] else 0 for i, gt in enumerate(gts)]
        return np.array(acc)

    def col_name(self) -> str:
        return "Hit-Rate"



import uptrain.v0 as v0
from test_helpers import KpsDataset, get_data_from_remote, read_json
import torch
from test_helpers import (
    pushup_signal,
    train_model_torch,
    get_accuracy_torch,
    BinaryClassification,
)


# if __name__ == "__main__":
def test_edge_cases():
    get_data_from_remote()

    training_file = "data/training_data.json"
    golden_testing_file = "data/golden_testing_data.json"

    # Real world test cases contains the data-points which the models sees in production
    real_world_test_cases = "data/real_world_testing_data.json"

    # To annotate the collected data points, we extract the ground truth from the master annotation file
    # (this can also do something like schedule an annotation job on Mechanical turk or
    # integrate with your other annotation pipelines).
    annotation_args = {"master_file": "data/master_annotation_data.json"}

    # Log the collected data-points to a local folder (this can also be a SQL table, a data warehouse etc.).
    data_save_folder = "uptrain_smart_data__edge_cases"

    # Defining the pushup egde-case signal
    pushup_edge_case = v0.Signal("Pushup", pushup_signal)

    # Defining the model confidence edge-case signal
    # That is, identify model confidence <0.9 as an edge-case
    low_conf_edge_case = (
        v0.Signal(v0.ModelSignal.BINARY_ENTROPY_CONFIDENCE, is_model_signal=True) < 0.9
    )

    cfg = {
        # Define your signal to identify edge cases
        "checks": [
            {
                "type": v0.Monitor.EDGE_CASE,
                "signal_formulae": (pushup_edge_case | low_conf_edge_case),
            }
        ],
        # Connect training pipeline to annotate data and retrain the model
        "training_args": {
            "annotation_method": {
                "method": v0.AnnotationMethod.MASTER_FILE,
                "args": annotation_args,
            },
            "training_func": train_model_torch,
            "orig_training_file": training_file,
        },
        # Retrain once 250 edge cases are collected
        "retrain_after": 250,
        # A local folder to store the retraining dataset
        "retraining_folder": "uptrain_smart_data_edge_cases_torch",
        # Connect evaluation pipeline to test retrained model against original model
        "evaluation_args": {
            "inference_func": get_accuracy_torch,
            "golden_testing_dataset": golden_testing_file,
        },
    }

    framework = v0.Framework(cfg)

    inference_batch_size = 16
    model_dir = "trained_models_torch/"
    model_save_name = "version_0"
    real_world_dataset = KpsDataset(
        real_world_test_cases,
        batch_size=inference_batch_size,
        shuffle=False,
        augmentations=False,
        is_test=True,
    )
    model = BinaryClassification()
    model.load_state_dict(torch.load(model_dir + model_save_name))
    model.eval()
    gt_data = read_json(annotation_args["master_file"])
    all_gt_ids = [x["id"] for x in gt_data]

    for i, elem in enumerate(real_world_dataset):
        # Do model prediction
        inputs = {"kps": elem[0]["kps"], "id": elem[0]["id"]}
        x_test = torch.tensor(inputs["kps"]).type(torch.float)
        test_logits = model(x_test).squeeze()
        preds = torch.round(torch.sigmoid(test_logits)).detach().numpy()
        idens = framework.log(inputs=inputs, outputs=preds)

        # Attach ground truth
        this_elem_gt = [gt_data[all_gt_ids.index(x)]["gt"] for x in elem[0]["id"]]
        framework.log(identifiers=idens, gts=this_elem_gt)

        # Retrain only once
        if framework.version > 1:
            break



import numpy as np
from uptrain.v0.core.lib.helper_funcs import fn_dependency_required

try:
    import faster_whisper
except:
    faster_whisper = None

try:
    import rouge
except:
    rouge = None

@fn_dependency_required(faster_whisper, "faster-whisper")
def faster_whisper_speech_to_text(audio_files):
    model_size = "large-v2"
    model = faster_whisper.WhisperModel(model_size, device="cpu", compute_type="int8")
    prescribed_texts = []
    for audio_file in audio_files:
        segments, _ = model.transcribe(audio_file, beam_size=5)
        prescribed_text = ''
        for segment in segments:
            prescribed_text += segment.text
        prescribed_texts.append(prescribed_text)
    return prescribed_texts

@fn_dependency_required(rouge, "rouge")
def rogue_l_similarity(text1_list, text2_list):
    r = rouge.Rouge()
    res = r.get_scores([x.lower() for x in text1_list],[x.lower() for x in text2_list])
    return np.array([x['rouge-l']['f'] for x in res])



import os
import subprocess

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
import uptrain.v0 as v0
from xgboost import XGBClassifier


def test_concept_drift_custom_monitor():
    data_file = "NSL_KDD_binary.csv"
    remote_url = "https://oodles-dev-training-data.s3.amazonaws.com/NSL_KDD_binary.csv"

    if not os.path.exists(data_file):
        try:
            # Most Linux distributions have Wget installed by default.
            # Below command is to install wget for MacOS
            wget_installed_ok = subprocess.call(
                "brew install wget",
                shell=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.STDOUT,
            )
            print("Successfully installed wget")
        except:
            pass

        try:
            if not os.path.exists("data.zip"):
                file_downloaded_ok = subprocess.call(
                    "wget " + remote_url,
                    shell=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.STDOUT,
                )
                print("Data downloaded")
        except:
            print("Could not load training data")
            print("Please follow following steps to manually download data")
            print(
                "Step 1: Open in browser: https://oodles-dev-training-data.s3.amazonaws.com/NSL_KDD_binary.csv"
            )
            print(
                "Step 2: Download and move the file to example location (i.e. uptrain/examples/2_cyber_attack_classification/"
            )
    else:
        print("Data file exists. Skipping download.")

    df = pd.read_csv(data_file)
    X = df.iloc[:, :-1].values
    y = df.iloc[:, -1].values

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, train_size=0.1, test_size=0.9, random_state=0, shuffle=False
    )

    # Train the XGBoost classifier with training data
    classifier = XGBClassifier()
    classifier.fit(X_train, y_train)

    y_pred = classifier.predict(X_train)
    print("Training accuracy: " + str(100 * accuracy_score(y_train, y_pred)))

    """
    Defining a custom drift metric where
    the user just want to check if accuracy
    drops beyond a threshold.
    """

    def custom_initialize_func(self):
        self.initial_acc = None
        self.acc_arr = []
        self.count = 0
        self.thres = 0.02
        self.window_size = 200
        self.is_drift_detected = False

    def custom_check_func(self, inputs, outputs, gts=None, extra_args={}):
        batch_size = len(extra_args["id"])
        self.count += batch_size
        self.acc_arr.extend(list(np.equal(gts, outputs)))

        # Calculate initial performance of the model on first 200 points
        if (self.count >= self.window_size) and (self.initial_acc is None):
            self.initial_acc = (
                sum(self.acc_arr[0 : self.window_size]) / self.window_size
            )

        # Calculate the most recent accuracy and log it to dashboard.
        if self.initial_acc is not None:
            for i in range(self.count - batch_size, self.count, self.window_size):
                # Calculate the most recent accuracy
                recent_acc = (
                    sum(self.acc_arr[i : i + self.window_size]) / self.window_size
                )

                # Logging to UpTrain dashboard
                self.log_handler.add_scalars(
                    "custom_metrics",
                    {
                        "y_initial_acc": self.initial_acc,
                        "y_recent_acc": recent_acc,
                    },
                    i,
                    self.dashboard_name,
                )

                # Send an alert when recent model performance goes down
                if (self.initial_acc - recent_acc > self.thres) and (
                    not self.is_drift_detected
                ):
                    alert = f"Concept drift detected with custom metric at time: {i}"
                    print(alert)
                    self.log_handler.add_alert(
                        "Custom Concept Drift Alert", alert, self.dashboard_name
                    )
                    self.is_drift_detected = True

    checks = [
        {
            "type": v0.Monitor.CUSTOM_MONITOR,
            "initialize_func": custom_initialize_func,
            "check_func": custom_check_func,
            "need_gt": True,
        },
    ]

    cfg = {
        # Checks to identify concept drift
        "checks": checks,
        # Folder that stores data logged by UpTrain
        "retraining_folder": "uptrain_smart_data",
    }

    # Initialize the UpTrain framework
    framework = v0.Framework(cfg)
    batch_size = len(X_test)

    for i in range(int(len(X_test) / batch_size)):
        # Do model prediction
        inputs = {"feats": X_test[i * batch_size : (i + 1) * batch_size]}
        preds = classifier.predict(inputs["feats"])

        # Log model inputs and outputs to monitor concept drift
        ids = framework.log(inputs=inputs, outputs=preds)

        # Attach ground truth to corresponding predictions
        # in UpTrain framework and identify concept drift
        ground_truth = y_test[i * batch_size : (i + 1) * batch_size]
        framework.log(identifiers=ids, gts=ground_truth)



"""
Implement checks to test language quality on different aspects. 

This module provides the `Critique` class, which evaluates a text generation on multiple 
aspects - fluence, politeness, grammar, and coherence,. It provides a score for each of the 
aspects on a scale of 0 to 1, along with an explanation for the score. 
"""

from __future__ import annotations
import typing as t

from loguru import logger
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *


@register_op
class LanguageCritique(ColumnOp):
    """
    Operator to score machine generated responses in a conversation. The response
    is evaluated on multiple aspects - fluence, politeness, grammar, and coherence.
    It provides a score for each of the aspects on a scale of 0 to 1, along with an
    explanation for the score.

    Attributes:
        col_response (str): The name of the input column containing response text
    
    Raises:
        Exception: Raises exception for any failed evaluation attempts
    
    """

    col_response: str = "response"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {"text": row[self.col_response]}
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate("critique_language", data_send)
        except Exception as e:
            logger.error(f"Failed to run evaluation for `LanguageCritique`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results))}


@register_op
class ToneCritique(ColumnOp):
    """
    Operator to assess the tone of machine generated responses.

    Attributes:
        llm_persona (str): The persona the chatbot being assessed was expected to follow
        col_response (str): The name of the input column containing response text
    
    Raises:
        Exception: Raises exception for any failed evaluation attempts
    
    """

    llm_persona: str = "helpful-chatbot"
    col_response: str = "response"
    col_out: str = "score_tone"

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {"response": row[self.col_response]}
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate(
                "critique_tone", data_send, {"llm_persona": self.llm_persona}
            )
        except Exception as e:
            logger.error(f"Failed to run evaluation for `ToneCritique`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({"score_tone": self.col_out}))}



from uptrain.v0.core.classes import Framework
from uptrain.v0.core.classes.signals import Signal
from uptrain.v0.core.classes.monitors import CustomMonitor
from uptrain.v0.constants import (
    AnnotationMethod,
    BiasAlgo,
    ClusteringAlgorithm,
    DataDriftAlgo,
    MeasurableType,
    ModelSignal,
    Monitor,
    PlotType,
    Statistic,
    Visual,
    ComparisonMetric,
    ComparisonModel,
)
from uptrain.v0.core.encoders import UpTrainEncoder



import re
from uptrain.v0.core.lib.helper_funcs import read_json, dependency_required
from uptrain.v0.core.classes.measurables import Measurable

try:
    import openai
except:
    openai = None


def extractNumber(text):
    return int(re.findall(r'\d+', text)[0])


@dependency_required(openai, "openai")
class GrammerScoreMeasurable(Measurable):
    def __init__(self, framework, feature_name) -> None:
        super().__init__(framework)
        self.feature_name = feature_name
        openai_key = framework.config_obj.license_args.openai_key
        if openai_key:
            openai.api_key = openai_key
        else:
            raise Exception("OpenAI key not found in config")
        self.client = openai.OpenAI()

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None):
        if self.feature_name:
            vals = inputs[self.feature_name]
        else:
            vals = outputs
        vals = [self.getGrammaticalCorrectnessScore(x) for x in vals]
        return vals

    def col_name(self) -> str:
        name = 'grammar_feature'
        if self.feature_name:
            name = name + '_' + self.feature_name
        return name
    
    # TODO: Decommission and find a generic way
    def extract_val_from_training_data(self, x):
        vals = x[self.feature_name]
        vals = [self.getGrammaticalCorrectnessScore(y) for y in vals]
        return vals
    
    def getGrammaticalCorrectnessScore(self, summary):
        response = self.client.chat.completions.create(model="gpt-3.5-turbo",
            messages=[
                {"role": "system", "content": "You are a grammatical correctnes evaluator who only gives only a number and no explanation."},
                {"role": "user", "content": "Score following sentence on grammatical correctness on a scale of 0 to 100: \n\n {statement}".format(statement=summary)},
            ],
            temperature=0,
        )
        try:
            return extractNumber(response.choices[0]['message']['content'])
        except:
            return -1




from .abstract_distance import AbstractDistance
from .cosine_distance import CosineDistance
from .l2_distance import L2Distance
from .norm_ratio import NormRatio

from .distance_resolver import DistanceResolver



import enum
import pydantic
import typing as t

class Evals(enum.Enum):
    CONTEXT_RELEVANCE = "context_relevance"
    FACTUAL_ACCURACY = "factual_accuracy"
    RESPONSE_RELEVANCE = "response_relevance"
    CRITIQUE_LANGUAGE = "critique_language"
    RESPONSE_COMPLETENESS = "response_completeness"
    RESPONSE_COMPLETENESS_WRT_CONTEXT = "response_completeness_wrt_context"
    RESPONSE_CONSISTENCY = "response_consistency"
    RESPONSE_CONCISENESS = "response_conciseness"
    VALID_RESPONSE = "valid_response"
    RESPONSE_ALIGNMENT_WITH_SCENARIO = "response_alignment_with_scenario"
    RESPONSE_SINCERITY_WITH_SCENARIO = "response_sincerity_with_scenario"
    PROMPT_INJECTION = "prompt_injection"


class ParametricEval(pydantic.BaseModel):
    ...


class CritiqueTone(ParametricEval):
    llm_persona: str = "helpful-chatbot"


class GuidelineAdherence(ParametricEval):
    guideline: str
    guideline_name: str = "guideline"  # User-assigned name of the guideline to distinguish between multiple checks
    response_schema: t.Union[str, None] = None  # Schema of the response in case it is of type JSON, XML, etc.

class ConversationSatisfaction(ParametricEval):
    user_persona: str = "user"
    llm_persona: t.Union[str, None] = None
    

class ResponseMatching(ParametricEval):
    method: str = "llm"



import uptrain.v0 as v0
import random
import numpy as np


# if __name__ == "__main__":
def test_dashboard():
    cfg = {
        "logging_args": {
            # For slack alerts, add your webhook URL
            # Checkout https://api.slack.com/messaging/webhooks
            'slack_webhook_url': None,
            'dashboard_port': 50000,
            'st_logging':True
        }
    }
    fw = v0.Framework(cfg)

    dashboard_name = "count_dashboard"
    for count in [0, 50, 100]:
        data = [count + random.randint(1, 100) for _ in range(100)]
        fw.log_handler.add_histogram(
            "multiple_histograms",
            data,
            dashboard_name,
            file_name="random(1,100) + " + str(count)
        )

    data = np.random.normal(size=1000)
    fw.log_handler.add_histogram(
        "single_histogram",
        data,
        dashboard_name,
    )

    fw.log_handler.add_alert(
        "Alerting",
        "A test alert",
        dashboard_name
    )

    line1 = [i + random.randint(1, 10) for i in range(100)]
    line2 = [i + random.randint(10, 20) for i in range(100)]

    for i,_ in enumerate(line1):
        fw.log_handler.add_scalars(
            "2 lines",
            {"y_value": line1[i]},
            i,
            dashboard_name,
            file_name='y_1',
        )
        fw.log_handler.add_scalars(
            "2 lines",
            {"y_value": line2[i]},
            i,
            dashboard_name,
            file_name='y_2',
        )
        fw.log_handler.add_scalars(
            "1 line",
            {"y_value": line2[i]+10},
            i,
            dashboard_name,
        )
        

    bar1_y = [i for i in range(5)]
    bar1_x = ["date: " + str(i) for i in range(5)]
    bar1_dict = dict(zip(bar1_x, bar1_y))

    bar2_y = [i for i in range(5,-1,-1)]
    bar2_x = ["date: " + str(i) for i in range(5)]
    bar2_dict = dict(zip(bar2_x, bar2_y))

    bar_dict = {'bar1': bar1_dict, 'bar2': bar2_dict}

    fw.log_handler.add_bar_graphs(
        "random bars",
        bar_dict,
        dashboard_name,
    )

    ########### Adding UMAP and TSNE test #############

    size = 300
    clusters = np.random.choice([0, 1], size=size, p=[0.2, 0.8])
    high_dim_list = []
    for label in clusters:
        if label==0:
            high_dim_list.append(np.random.normal(np.zeros(size), 1, size))
        else:
            high_dim_list.append(np.random.normal(np.ones(size), 1, size))

    config_umap = {
        "checks": [
        {
            'type': v0.Visual.UMAP,
            "measurable_args": {
                'type': v0.MeasurableType.INPUT_FEATURE,
                'feature_name': 'data'
            },
            "label_args": [{
                'type': v0.MeasurableType.INPUT_FEATURE,
                'feature_name': 'labels'
            }],
            'min_dist': 0.01,
            'n_neighbors': 20,
            'metric': 'euclidean',
            'dim': '2D',
            "update_freq": 1,
        },
        {
            'type': v0.Visual.TSNE,
            "measurable_args": {
                'type': v0.MeasurableType.INPUT_FEATURE,
                'feature_name': 'data'
            },
            "label_args": [{
                'type': v0.MeasurableType.INPUT_FEATURE,
                'feature_name': 'labels'
            }],
            'dim': '3D',
            "update_freq": 1,
            'perplexity': 10,
        }
        ],
        "logging_args": {
            'log_folder': 'uptrain_logs_umap',
            'dashboard_port': 50001,
            'st_logging': True,

        }
    }

    framework_umap = v0.Framework(cfg_dict=config_umap)

    inputs = {'data': high_dim_list, 'labels': clusters}
    idens = framework_umap.log(inputs=inputs)

    


"""
Implement operators to generate text. Used when evaluating different 
prompts/LLM-input-parameters for gnerating text. 
"""

from __future__ import annotations
import itertools
import typing as t
import json
import numpy as np 

from loguru import logger
import polars as pl
from uptrain.framework import Settings

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.operators.language.llm import LLMMulticlient, Payload


@register_op
class PromptGenerator(TransformOp):
    """Operator to generate text given different prompts/LLM-input-parameters.

    Attributes:
        prompt_template (str) : A string template for the prompt.
        prompt_params (dict[str, list[str]]): A dictionary mapping parameter names to lists of values.
            The cartesian product of all the parameter values will be used to
            construct the prompts.
        models (list[str]): A list of models to run the experiment on.
        context_vars (Union[list[str], dict[str, str]]): A dictionary mapping context variable names to corresponding
            columns in the input dataset.
        col_out_prefix (str): Prefix for the output columns.

    """

    prompt_template: str
    prompt_params: dict[str, list[str]]
    models: list[str]
    context_vars: t.Union[list[str], dict[str, str]]
    col_out_prefix: str = "exp_"

    def setup(self, settings: "Settings"):
        self._settings = settings
        if isinstance(self.context_vars, list):
            self.context_vars = dict(zip(self.context_vars, self.context_vars))
        return self

    """Construct all the prompt variations and generate completions for each."""

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        list_params = []
        for experiment_id, combo in enumerate(
            itertools.product(*self.prompt_params.values(), self.models)
        ):
            prompt_params, model = combo[:-1], combo[-1]
            variables = dict(zip(self.prompt_params.keys(), prompt_params))
            list_params.append(
                {
                    "template": self.prompt_template,
                    self.col_out_prefix + "model": model,
                    **{self.col_out_prefix + k: v for k, v in variables.items()},
                    self.col_out_prefix + "experiment_id": experiment_id,
                }
            )
        params_dataset = pl.DataFrame(list_params)

        # Do a cross join of the input with the params dataset to get all the
        # inputs for the completion step
        input_dataset = data.join(params_dataset, on=None, how="cross")

        # construct the prompts by iterating over the rows of the input dataset and
        # formatting the prompt template with the row values
        prompts = []
        for row in input_dataset.iter_rows(named=True):
            fill = {k: row[v] for k, v in self.context_vars.items()}
            fill.update(
                {k: row[self.col_out_prefix + k] for k in self.prompt_params.keys()}
            )

            # TODO: Temp Fix for handling json in prompts. Permanent fix is to integrate langchain?
            try:
                prompt = row["template"].format(**fill)
            except:
                prompt = row["template"]
                for k, v in fill.items():
                    prompt = prompt.replace("{{" + k + "}}", v)
            prompts.append(prompt)

        input_w_prompts = input_dataset.with_columns(
            [pl.Series(self.col_out_prefix + "prompt", prompts)]
        )
        return {"output": input_w_prompts}

        # # get text completions for each row
        # op_completion = TextCompletion(
        #     col_in_prompt="exp_prompt", col_in_model="exp_model"
        # )
        # completions = op_completion.setup(self._settings).run(input_w_prompts)["output"]
        # assert completions is not None
        # return {
        #     "output": input_w_prompts.with_columns([completions.alias("exp_generated")])
        # }


@register_op
class TextCompletion(TransformOp):
    """
    Takes a table of prompts and LLM model to use, generates output text.

    Attributes:
        col_in_prompt (str): The name of the column containing the prompt template.
        col_in_model (str): The name of the column containing the model name.
        col_out_completion (str): The name of the column containing the generated text.
        temperature (float): Temperature for the LLM to generate responses.

    Returns:
        TYPE_TABLE_OUTPUT: A dictionary containing the dataset with the output text.
    """

    col_in_prompt: str = "prompt"
    col_in_model: str = "model"
    col_out_completion: str = "generated"
    temperature: float = 1.0
    _api_client: LLMMulticlient

    def setup(self, settings: Settings):
        self._api_client = LLMMulticlient(settings=settings)
        self._settings = settings
        return self

    def _make_payload(self, id: t.Any, text: str, model: str) -> Payload:
        if self._settings.seed is not None:
            return Payload(
                data={
                    "model": model,
                    "messages": [{"role": "user", "content": text}],
                    "temperature": self.temperature,
                    "seed" : self._settings.seed
                },
                metadata={"index": id},
            )
        else:
            return Payload(
                data={
                    "model": model,
                    "messages": [{"role": "user", "content": text}],
                    "temperature": self.temperature
                },
                metadata={"index": id},
            )

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        prompt_ser = data.get_column(self.col_in_prompt)
        model_ser = data.get_column(self.col_in_model)
        input_payloads = [
            self._make_payload(idx, text, model)
            for idx, (text, model) in enumerate(zip(prompt_ser, model_ser))
        ]
        output_payloads = self._api_client.fetch_responses(input_payloads)

        results = []
        for res in output_payloads:
            assert (
                res is not None
            ), "Response should not be None, we should've handled exceptions beforehand."
            idx = res.metadata["index"]
            if res.error is not None:
                logger.error(
                    f"Error when processing payload at index {idx}: {res.error}"
                )
                results.append((idx, None))
            else:
                resp_text = res.response.choices[0].message.content
                results.append((idx, resp_text))

        output_text = pl.Series(
            values=[val for _, val in sorted(results, key=lambda x: x[0])]
        )
        return {
            "output": data.with_columns([output_text.alias(self.col_out_completion)])
        }


@register_op
class TopicGenerator(ColumnOp):
    """
    Takes a table of clustered texts and identifies the topic for each cluster 

    Attributes:
        col_in_cluster_index (str): The name of the column containing the cluster index.
        col_in_dist (str): The name of the column containing the euclidean distance from its cluster centroid.
        col_in_text (str): The name of the column containing the text where the grouping needs to be performed.
        top_n (int): Number of examples to be considered for each category. 
        col_out_text (str): The name of the column containing the topic for each entry.
        temperature (float): Temperature for the LLM to generate responses.

    Returns:
        TYPE_TABLE_OUTPUT: A dictionary containing the dataset with the output text.
    """
    col_in_cluster_index: str = 'cluster_index'
    col_in_dist: str = 'cluster_index_distance'
    col_in_text: str = 'question'
    top_n: int = 5 
    col_out_text: str = 'topic'
    temperature: float = 1.0
    _api_client: LLMMulticlient

    def setup(self, settings: Settings):
        self._api_client = LLMMulticlient(settings=settings)
        self._settings = settings
        self.model = settings.model
        return self

    def _make_payload(self, id: t.Any, text: str, model: str) -> Payload:
        if self._settings.seed is not None:
            return Payload(
                data={
                    "model": model,
                    "messages": [{"role": "user", "content": text}],
                    "temperature": self.temperature,
                    "seed" : self._settings.seed
                },
                metadata={"index": id},
            )
        else:
            return Payload(
                data={
                    "model": model,
                    "messages": [{"role": "user", "content": text}],
                    "temperature": self.temperature
                },
                metadata={"index": id},
            )

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:

        res_data_arr = []

        unique_agg_keys = ['default']
        if '_unique_agg_key_for_clustering' in list(data.columns):
            # First aggregate by col_aggs
            agg_data = data.groupby('_unique_agg_key_for_clustering').agg([pl.col(self.col_in_text).count().alias("num_rows_" + self.col_in_text)])
            agg_data = agg_data.drop("num_rows_" + self.col_in_text)
            unique_agg_keys = [eval(x) for x in list(agg_data['_unique_agg_key_for_clustering'])]

        self.topics = {}

        for unique_agg_key in unique_agg_keys:
            cond = True
            if isinstance(unique_agg_key, dict):
                unique_agg_key = dict([(key, unique_agg_key[key]) for key in sorted(unique_agg_key)])
                for key,val in unique_agg_key.items():
                    cond = cond & (data[key] == val)
            data_subset = data.filter(cond)

            questions = np.asarray(data_subset[self.col_in_text])
            distances = np.asarray(data_subset[self.col_in_dist])

            n_clusters = max(data_subset[self.col_in_cluster_index])
            input_payloads = []
            outputs = [None] * len(questions)
            indexes_cluster = []

            for index in range(n_clusters+1):
                points = data_subset[self.col_in_cluster_index]==index
                indexes_cluster.append(list(np.where(points)[0]))
                distances_cluster = distances[points]
                questions_cluster = questions[points]
                

                indexes_sorted = np.argsort(distances_cluster)
                indexes_top_n = indexes_sorted[:min(self.top_n, len(questions_cluster))]
                questions_top_n = questions_cluster[indexes_top_n]

                text = ''
                for j in range(len(questions_top_n)):
                    text = text + str(j+1) + '. ' + str(questions_top_n[j]) + '\n'
                
                input = f"""{text}"""
                prompt = f""" Identify the common topic in the given sentences below. 
                {input}
                
                You should identify only a single concise topic which should be at max 15 words long. Just print the topic and nothing else. 
                Topic: """
                input_payloads.append(self._make_payload(index, prompt, self.model))
        

            output_payloads = self._api_client.fetch_responses(input_payloads)

            results = []
            for res in output_payloads:
                assert (
                    res is not None
                ), "Response should not be None, we should've handled exceptions beforehand."
                idx = res.metadata["index"]
                if res.error is not None:
                    logger.error(
                        f"Error when processing payload at index {idx}: {res.error}"
                    )
                    results.append((idx, None))
                else:
                    resp_text = res.response.choices[0].message.content
                    results.append((idx, resp_text))
            
            for index, resp_text in results:
                indexes = indexes_cluster[index]
                for idx in indexes:
                    outputs[idx]=resp_text

            output_text = pl.Series(
                values=outputs
            )
            data_subset = data_subset.with_columns([output_text.alias(self.col_out_text)])
            res_data_arr.append(data_subset)

            results.sort(key=lambda x: x[0])
            self.topics[str(unique_agg_key)] = [x[1] for x in results]

        return {
            "output": pl.concat(res_data_arr)
        }


@register_op
class OutputParser(ColumnOp):
    """
    Takes a table of LLM respones and parses them into individual columns

    Attributes:
        col_in_response (str): The name of the column containing the raw model response.
        col_out_mapping (str): A dictionary containing the mapping of keys in response to their output column names.

    Returns:
        TYPE_TABLE_OUTPUT: A dictionary containing the dataset with the output text.
    """

    col_in_response: str
    col_out_mapping: dict

    def setup(self, settings: "Settings"):
        return self

    """Parse the LLM responses"""

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        responses = data[self.col_in_response]
        parsed_responses = pl.DataFrame([json.loads(x) for x in responses])
        return {
            "output": data.with_columns(
                [parsed_responses[k].alias(v) for k, v in self.col_out_mapping.items()]
            )
        }



import numpy as np
import pandas as pd
from sklearn import preprocessing
import os
import subprocess

def process(data): 
    data["diff_lat"] = abs(data["pickup_latitude"] - data["dropoff_latitude"])
    data["diff_lon"] = abs(data["pickup_longitude"] - data["dropoff_longitude"])
    
    data["dist"] = np.sqrt(np.power(data["diff_lat"],2) + np.power(data["diff_lon"],2))

    data["pickup_datetime"] = pd.to_datetime(data["pickup_datetime"])
    
    data["pickup_year"] = data["pickup_datetime"].dt.year
    data["pickup_month"] = data["pickup_datetime"].dt.month
    data["pickup_day"] = data["pickup_datetime"].dt.day
    data["pickup_dow"] = data["pickup_datetime"].dt.dayofweek
    data["pickup_hour"] = data["pickup_datetime"].dt.hour
    data["pickup_min"] = data["pickup_datetime"].dt.minute
    data["pickup_sec"] = data["pickup_datetime"].dt.second
    
    data = data.drop(["diff_lat", "diff_lon", "pickup_longitude", "pickup_latitude", "dropoff_latitude", "dropoff_longitude", "id", "pickup_datetime"], axis = 1)
    data = data.drop(["dropoff_datetime"], axis = 1)

    le = preprocessing.LabelEncoder()
    data["store_and_fwd_flag"] = le.fit_transform(data["store_and_fwd_flag"])
    return data


def download_dataset(data_file):    
    remote_url = "https://oodles-dev-training-data.s3.amazonaws.com/trip_duration_dataset.csv"

    if not os.path.exists(data_file):
        print("Installing wget to download dataset from remote")
        try:
            # Most Linux distributions have Wget installed by default.
            # Below command is to install wget for MacOS
            wget_installed_ok = subprocess.call("brew install wget", shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
            print("Successfully installed wget")
        except:
            print("Wget installation fails! Checking if data is manually downloaded")
            dummy = 1
        try:
            print("Downloading data from the remote server")
            file_downloaded_ok = subprocess.call("wget " + remote_url, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
            print("Data downloaded")
        except Exception as e:
            print(e)
            print("Could not load training data")
            print("Please follow following steps to manually download data")
            print(f"Step 1: Open in browser: {remote_url}")
            print("Step 2: Download and move the file to the example location (i.e. uptrain/examples/ride_time_estimation/")
    else:
        print("Data file exists. Skipping download.")



"""
Implement checks to test if a piece of text has been taken from a source.

This module provides the `RougeScore` class, which allows comparing a generated text with a source text to check if it has been taken from the source. It uses the Rouge score metric, specifically the Rouge-L score, to measure the similarity between the two texts.

"""

from __future__ import annotations
import typing as t

from loguru import logger
import polars as pl
from uptrain.framework import Settings

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import lazy_load_dep

rouge_scorer = lazy_load_dep("rouge_score.rouge_scorer", "rouge_score")


@register_op
class RougeScore(ColumnOp):
    """
    Operator to compare a generated text with a source text using the Rouge score metric.

    Attributes:
        score_type (Literal["precision", "recall", "f1"]): The type of Rouge score to calculate.
        col_in_generated (str): The name of the input column containing the generated text.
        col_in_source (str): The name of the input column containing the source text.
        col_out (str): The name of the output column containing the Rouge scores.

    Returns:
        dict: A dictionary containing the Rouge scores for each pair of generated and source text.

    Example:
        ```
        import polars as pl
        from uptrain.operators import RougeScore

        # Create a DataFrame
        df = pl.DataFrame({
            "text_generated": ["This is the generated text.", "Another generated sentence."],
            "text_source": ["This is the original source text.", "This is a different source text."]
        })

        # Create an instance of the RougeScore class
        rouge_op = RougeScore(score_type="f1")

        # Calculate the Rouge-L scores
        scores = rouge_op.run(df)["output"]

        # Print the Rouge-L scores
        print(scores)
        ```

    Output:
        ```
        shape: (2,)
        Series: '_col_0' [i64]
        [
                72
                0
        ]
        ```

    """

    score_type: t.Literal["precision", "recall", "f1"]
    col_in_generated: str = "text_generated"
    col_in_source: str = "text_source"
    col_out: str = "rouge_score"

    def setup(self, settings: Settings):
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        text_generated = data.get_column(self.col_in_generated)
        text_source = data.get_column(self.col_in_source)

        results = []
        scores = []
        for i in range(len(text_generated)):
            scorer = rouge_scorer.RougeScorer(["rougeL"])  # type: ignore
            if text_source[i] is None or text_generated[i] is None:
                scores.append({"rougeL": (0, 0, 0)})
            else:
                scores.append(scorer.score(text_source[i], text_generated[i]))

        type_to_index = {"precision": 0, "recall": 1, "f1": 2}
        if self.score_type not in type_to_index:
            raise Exception(f"{self.score_type} not implemented")
        else:
            score_index = type_to_index[self.score_type]

        results = pl.Series([int(x["rougeL"][score_index] * 100) for x in scores])
        return {"output": data.with_columns([results.alias(self.col_out)])}



"""
Implement checks to test language quality. 
"""

from __future__ import annotations
import asyncio
from concurrent.futures import ThreadPoolExecutor
import random
import typing as t

from loguru import logger
from pydantic import BaseModel, Field

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import lazy_load_dep

openai = lazy_load_dep("openai", "openai")
litellm = lazy_load_dep("litellm", "litellm")
aiolimiter = lazy_load_dep("aiolimiter", "aiolimiter>=1.1")
tqdm_asyncio = lazy_load_dep("tqdm.asyncio", "tqdm>=4.0")


from openai import AsyncOpenAI
from openai import AsyncAzureOpenAI
import openai
#import openai.error


# -----------------------------------------------------------
# Make concurrent requests to the OpenAI or another LLM API
#
# Credits: https://github.com/cozodb/openai-multi-client/
# Seems to get stuck for me though.
# And zenoml code - https://github.com/zeno-ml/zeno-build/blob/main/zeno_build/models/providers/openai_utils.py
# -----------------------------------------------------------


class Payload(BaseModel):
    data: dict
    metadata: dict = Field(default_factory=dict)
    response: t.Any = None
    error: t.Optional[str] = None


async def async_process_payload(
    payload: Payload,
    rpm_limiter: aiolimiter.AsyncLimiter,
    tpm_limiter: aiolimiter.AsyncLimiter,
    aclient: t.Union[AsyncOpenAI, AsyncAzureOpenAI, None],
    max_retries: int,
) -> Payload:
    messages = payload.data["messages"]
    total_chars = sum(len(msg["role"]) + len(msg["content"]) for msg in messages)
    total_tokens = total_chars // 3  # average token length is 3, conservatively
    await rpm_limiter.acquire(1)
    # TODO: we should also count the response tokens, but this is a good baseline
    # since our use-case is evaluations mostly, not generation
    await tpm_limiter.acquire(total_tokens)

    for count in range(max_retries):  # failed requests don't count towards rate limit
        try:
            if aclient is not None:
                payload.response = await aclient.chat.completions.create(**payload.data, timeout=180)
            else:
                payload.response = await litellm.acompletion(
                    **payload.data,
                )
            break
        except Exception as exc:
            logger.error(f"Error when sending request to LLM API: {exc}")
            if (
                isinstance(
                    exc,
                    (
                        litellm.RateLimitError,
                        litellm.llms.azure.AzureOpenAIError,
                        openai.APIConnectionError,
                        openai.APITimeoutError,
                        openai.RateLimitError,
                        openai.InternalServerError,
                        openai.APIError,
                        openai.Timeout
                    ),
                )
                and count < max_retries - 1
            ):
                await asyncio.sleep(random.uniform(5, 30) * count + 60)
            elif (
                isinstance(exc, openai.BadRequestError)
                and "context_length" in exc.code
                and count < max_retries - 1
            ):
                # refer - https://github.com/BerriAI/reliableGPT/
                # if required to set token limit - https://learn.microsoft.com/en-us/azure/ai-services/openai/how-to/chatgpt?pivots=programming-language-chat-completions#managing-conversations

                fallback = {
                    "gpt-3.5-turbo": "gpt-3.5-turbo-16k",
                    "gpt-3.5-turbo-0613": "gpt-3.5-turbo-16k-0613",
                    "gpt-3.5-turbo-16k": "claude-instant-1.2",
                    "gpt-3.5-turbo-16k-0613": "claude-instant-1.2",
                    "gpt-4": "gpt-4-32k",
                    "gpt-4-0613": "gpt-4-32k-0613",
                    "gpt-4-32k": "claude-2.0",
                    "gpt-4-32k-0613": "claude-2.0",
                }
                if payload.data["model"] in fallback.keys():
                    payload.data["model"] = fallback[payload.data["model"]]
                    logger.info(
                        f"Switching to larger context model for payload {payload.metadata['index']}"
                    )
            else:
                payload.error = str(exc)
                break

    return payload


class LLMMulticlient:
    """Uses asyncio to send requests to LLM APIs concurrently."""

    def __init__(self, settings: t.Optional[Settings] = None):
        self._max_tries = 4
        # TODO: consult for accurate limits - https://platform.openai.com/account/rate-limits
        self._rpm_limit = 200
        self._tpm_limit = 90_000
        self.aclient = None
        if settings is not None:
            if settings.openai_api_key is not None:
                openai.api_key = settings.check_and_get("openai_api_key")  # type: ignore
                self.aclient = AsyncOpenAI()

            if settings.model.startswith("azure") and settings.check_and_get("azure_api_key") is not None:
                self.aclient = AsyncAzureOpenAI(
                    api_key = settings.azure_api_key,  
                    api_version=settings.azure_api_version,
                    azure_endpoint = settings.azure_api_base
                )

            self._rpm_limit = settings.check_and_get("rpm_limit")
            self._tpm_limit = settings.check_and_get("tpm_limit")

    def fetch_responses(self, input_payloads: list[Payload]) -> list[Payload]:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        if loop and loop.is_running():
            logger.warning(
                "Detected a running event loop, scheduling requests in a separate thread."
            )
            with ThreadPoolExecutor(max_workers=1) as executor:
                return executor.submit(
                    asyncio.run, self.async_fetch_responses(input_payloads)
                ).result()
        else:
            return asyncio.run(self.async_fetch_responses(input_payloads))

    async def async_fetch_responses(
        self, input_payloads: list[Payload]
    ) -> list[Payload]:
        rpm_limiter = aiolimiter.AsyncLimiter(self._rpm_limit, time_period=60)
        tpm_limiter = aiolimiter.AsyncLimiter(self._tpm_limit, time_period=60)
        async_outputs = [
            async_process_payload(data, rpm_limiter, tpm_limiter, self.aclient, self._max_tries)
            for data in input_payloads
        ]
        output_payloads = await tqdm_asyncio.tqdm_asyncio.gather(*async_outputs)
        return output_payloads



import importlib.util
import os
import threading
from typing import Optional


class StreamlitRunner:
    log_folder: str
    launch_cmd: str

    def __init__(self, log_folder: str, port: Optional[int] = None):
        self.log_folder = log_folder

        path_uptrain_init: str = importlib.util.find_spec("uptrain").origin  # type: ignore
        path_st_run = os.path.join(
            os.path.dirname(path_uptrain_init), "v0/core/classes/logging/new_st_run.py"
        )
        port_arg = "" if port is None else f"--server.port {port}"
        self.launch_cmd = f"streamlit run {path_st_run} {port_arg} -- {self.log_folder}"

    def start(self):
        t = threading.Thread(target=lambda: os.system(self.launch_cmd), args=([]))
        t.start()



"""
Test that the builtin checks work without raising an error. NOT that they are appropriate for the task.
"""

import polars as pl

from uptrain.framework import CheckSet, Settings
from uptrain.framework.builtins import (
    CheckContextRelevance,
    CheckLanguageQuality,
    CheckResponseCompleteness,
    CheckResponseFacts,
    CheckToneQuality,
    CheckResponseRelevance
)

settings = Settings()
dataset = pl.DataFrame(
    {
        "response": [
            "The actress who played Lolita, Sue Lyon, was 14 at the time of filming.",
            "Shakespeare wrote 154 sonnets.",
            "Sachin Tendulkar retired from cricket in 2013.",
            "Python language was created by Guido van Rossum.",
            "The first manned Apollo mission was Apollo 1.",
        ],
        "question": [
            "What was the age of Sue Lyon when she played Lolita?",
            "How many sonnets did Shakespeare write?",
            "When did Sachin Tendulkar retire from cricket?",
            "Who created the Python language?",
            "Which was the first manned Apollo mission?",
        ],
        "context": [
            "Lolita is a 1962 psychological comedy-drama film directed by Stanley Kubrick. The film follows Humbert Humbert, a middle-aged literature lecturer who becomes infatuated with Dolores Haze, a young adolescent girl. It stars Sue Lyon as the titular character.",
            "William Shakespeare was an English playwright and poet, widely regarded as the world's greatest dramatist. He is often called the Bard of Avon. His works consist of some 39 plays, 154 sonnets and a few other verses.",
            "Sachin Tendulkar is a former international cricketer from India. He is widely regarded as one of the greatest batsmen in the history of cricket. He is the highest run scorer of all time in International cricket and played until 16 May 2013.",
            "Python is a high-level general-purpose programming language. Its design philosophy emphasizes code readability. Its language constructs aim to help programmers write clear, logical code for both small and large-scale projects.",
            "The Apollo program was a human spaceflight program carried out by NASA. It accomplished landing the first humans on the Moon from 1969 to 1972. The program was named after Apollo, the Greek god of light, music, and the sun. The first mission flown was dubbed as Apollo 1.",
        ],
    }
)


def test_check_context_relevance():
    check = CheckContextRelevance()
    output = check.setup(settings).run(dataset)
    assert isinstance(output, pl.DataFrame)


def test_check_language_quality():
    check = CheckLanguageQuality()
    output = check.setup(settings).run(dataset)
    assert isinstance(output, pl.DataFrame)


def test_check_response_completeness():
    check = CheckResponseCompleteness()
    output = check.setup(settings).run(dataset)
    assert isinstance(output, pl.DataFrame)


def test_check_response_relevance():
    check = CheckResponseRelevance()
    output = check.setup(settings).run(dataset)
    assert isinstance(output, pl.DataFrame)


def test_check_response_facts():
    check = CheckResponseFacts()
    output = check.setup(settings).run(dataset)
    assert isinstance(output, pl.DataFrame)


def test_check_tone_quality():
    check = CheckToneQuality(llm_persona="wikipedia-bot")
    output = check.setup(settings).run(dataset)
    assert isinstance(output, pl.DataFrame)



import numpy as np
import random
import uptrain.v0 as v0


def test_visuals_plot():
    """Test Visual Plot API functionality.
    
    In this test, we test the plotting functionality of the framework.
    We test the following plots:
        - Line chart
        - Bar chart
        - Histogram
    
    For each of the above plots, we generate random data.

    The test passes if the plots are generated correctly.
    """

    # Data for line chart
    line_n = 100
    epoch = list(range(line_n))
    loss = [random.randint(0, 100) for _ in range(line_n)]

    # Data for bar chart
    bar_n = 10
    bar1_y = [i for i in range(bar_n)]
    bar1_x = [f"feature: {i}" for i in range(bar_n)]
    bar2_y = [i for i in range(bar_n - 1, -1, -1)]
    bar2_x = [f"feature: {i}" for i in range(bar_n)]

    # Data for histogram
    hist_n = 1000
    hist_model_a = np.random.normal(size=hist_n)
    hist_model_b = np.random.gumbel(size=hist_n)

    # Create a configuration for the framework
    cfg = {
        # Define checks for the framework
        "checks": [
            # Line chart for epoch vs loss on default dashboard
            {
                "type": v0.Visual.PLOT,
                "plot": v0.PlotType.LINE_CHART,
                "x_feature_name": "epoch",
                "y_feature_name": "loss",
                "plot_name": "epoch vs loss",
            },

            # Line chart for epoch vs loss on a different dashboard
            {
                "type": v0.Visual.PLOT,
                "plot": v0.PlotType.LINE_CHART,
                "y_feature_name": "loss",
                "plot_name": "epoch vs loss 2",
                "dashboard_name": "Plot 2"
            },

            # Bar chart for bar1 and bar2 on default dashboard
            {
                "type": v0.Visual.PLOT,
                "plot": v0.PlotType.BAR_CHART,
                "x_feature_name": "key",
                "y_feature_name": "value",
                "bars": ["bar1", "bar2"],
                "plot_name": "bar chart",
            },

            # Histogram for model_a on default dashboard
            {
                "type": v0.Visual.PLOT,
                "plot": v0.PlotType.HISTOGRAM,
                "feature_name": "model_a",
                "plot_name": "single histogram",
            },

            # Histogram for model_b on default dashboard with multiple models
            {
                "type": v0.Visual.PLOT,
                "plot": v0.PlotType.HISTOGRAM,
                "feature_name": "model_a",
                "plot_name": "multiple histogram",
            },

            # Histogram for model_b on default dashboard with multiple models
            {
                "type": v0.Visual.PLOT,
                "plot": v0.PlotType.HISTOGRAM,
                "feature_name": "model_b",
                "plot_name": "multiple histogram",
            },
        ],

        # Specify logging arguments
        # "st_logging" should be True if we want streamlit logging, False otherwise
        "logging_args": {"st_logging": True},
    }

    framework = v0.Framework(cfg)

    # Log data to the framework
    framework.log({"epoch": epoch, "loss": loss})
    framework.log(
        {
            "bars": ["bar1"] * bar_n + ["bar2"] * bar_n,
            "key": bar1_x + bar2_x,
            "value": bar1_y + bar2_y,
        }
    )
    framework.log({"model_a": hist_model_a})
    framework.log({"model_b": hist_model_b})



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np

from uptrain.v0.core.classes.statistics import AbstractStatistic
from uptrain.v0.core.classes.measurables import MeasurableResolver
from uptrain.v0.constants import Statistic

if TYPE_CHECKING:
    from uptrain.v0.core.classes.logging.new_log_handler import (
        LogHandler as NewLogHandler,
        LogWriter,
    )
    from uptrain.v0.core.classes.logging.log_handler import LogHandler


class NormEmbedding(AbstractStatistic):
    """Class that computes the norm of an embedding column."""

    log_handler: "NewLogHandler"
    log_writer: "LogWriter"
    dashboard_name = "norm_embedding"
    statistic_type = Statistic.NORM_EMBEDDING

    def base_init(self, fw, check):
        self.aggregate_measurable = MeasurableResolver(check["aggregate_args"]).resolve(
            fw
        )
        self.count_measurable = MeasurableResolver(check["count_args"]).resolve(fw)

        # get handles to log writers for each distance type
        if hasattr(self.log_handler, "make_logger"):
            self.log_writer = self.log_handler.make_logger(self.dashboard_name, "check")
        else:
            raise Exception(
                f"Check type: {self.statistic_type} not supported by this log handler"
            )

    def base_check(self, inputs=None, outputs=None, gts=None, extra_args={}):
        all_models = [
            x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
            for x in self.model_measurables
        ]
        all_features = [
            x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
            for x in self.feature_measurables
        ]

        vals_all = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        aggregate_ids_all = self.aggregate_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        counts_all = self.count_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )

        if len(vals_all) == 0:
            return  # nothing for us to do here

        emb_values = np.vstack(vals_all)
        norm_values = np.linalg.norm(emb_values, axis=1)

        for idx in range(len(aggregate_ids_all)):
            models = dict(
                zip(
                    self.model_names,
                    [all_models[jdx][idx] for jdx in range(len(self.model_names))],
                )
            )
            features = dict(
                zip(
                    self.feature_names,
                    [all_features[jdx][idx] for jdx in range(len(self.feature_names))],
                )
            )
            self.log_writer.log(
                {
                    "check": norm_values[idx],
                    self.aggregate_measurable.feature_name: aggregate_ids_all[idx],  # type: ignore
                    self.count_measurable.feature_name: counts_all[idx],  # type: ignore
                    **models,
                    **features,
                }
            )



import os
import pickle

try:
    import shap
except:
    shap = None

from uptrain.v0.core.classes.visuals import AbstractVisual
from uptrain.v0.constants import Visual
from uptrain.v0.core.lib.helper_funcs import dependency_required


@dependency_required(shap, "shap")
class Shap(AbstractVisual):
    visual_type = Visual.SHAP
    dashboard_name = "SHAP_explanation"

    def base_init(self, fw, check):
        self.model = check["model"]
        self.explainer_created = False

        filename = "explainer.pkl"
        self.fileloc = os.path.join(fw.log_handler.st_log_folder, filename)
        feat_name_list = check.get("feat_name_list", fw.feat_name_list)

        fw.log_handler.add_st_metadata(
            {
                "path_all_data": fw.path_all_data,
                "feat_name_list": feat_name_list,
                "shap_num_points": check.get("shap_num_points", 1000),
                "path_shap_file": self.fileloc,
            }
        )

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        if not self.explainer_created:
            explainer = shap.Explainer(self.model)
            file = open(self.fileloc, "wb")
            pickle.dump(explainer, file)
            file.close()
            self.explainer_created = True



import numpy as np


def body_length_from_kps(kps):
    head_mean_point = np.sum(kps[0:5, 0:2], axis=0) / 5
    legs_mean_point = np.sum(kps[11:17, 0:2], axis=0) / 6
    body_length_horizontal = max(0.01, abs(head_mean_point[0] - legs_mean_point[0]))
    body_length_vertical = max(0.01, abs(head_mean_point[1] - legs_mean_point[1]))
    return max(body_length_horizontal, body_length_vertical)


def body_length_signal(inputs, outputs, gts=None, extra_args={}):
    body_lengths = []
    for input in inputs["kps"]:
        kps = np.reshape(np.array(input), (17, 2))
        body_length = body_length_from_kps(kps)
        body_lengths.append(body_length)
    return body_lengths


def pushup_signal(inputs, outputs, gts=None, extra_args={}):
    # Define signal to identify cases when the user is in the pushup position
    is_pushups = []
    for input in inputs["kps"]:
        kps = np.reshape(np.array(input), (17, 2))
        head_mean_point = np.sum(kps[0:5, 0:2], axis=0) / 5
        legs_mean_point = np.sum(kps[11:17, 0:2], axis=0) / 6
        body_slope = abs(
            (legs_mean_point[1] - head_mean_point[1])
            / max(0.1, abs(legs_mean_point[0] - head_mean_point[0]))
        )
        hands_mean_point = np.sum(kps[7:11, 0:2], axis=0) / 4
        shoulder_mean_point = np.sum(kps[5:7, 0:2], axis=0) / 2
        wrist_mean_point = np.sum(kps[9:11, 0:2], axis=0) / 2
        body_length = body_length_from_kps(kps)
        is_front_orientation = (kps[9, 0] - kps[10, 0]) > body_length
        is_pushup = (
            (body_slope < 1)
            and (hands_mean_point[1] > shoulder_mean_point[1])
            and (
                (
                    abs((wrist_mean_point[0] - shoulder_mean_point[0]) / body_length)
                    < 0.25
                )
                or is_front_orientation
            )
        )
        is_pushups.append(bool(is_pushup))
    return is_pushups



from .abstract_measurable import AbstractMeasurable
from .measurable import Measurable
from .input_feature import InputFeatureMeasurable
from .feature_concat import FeatureConcatMeasurable
from .output_feature import OutputFeatureMeasurable
from .feature import FeatureMeasurable
from .condition import ConditionMeasurable
from .custom import CustomMeasurable
from .accuracy import AccuracyMeasurable, MAEMeasurable, MAPEMeasurable
from .scalar_from_embedding import ScalarFromEmbeddingMeasurable
from .distance import DistanceMeasurable
from .recommendation_hit_rate import RecHitRateMeasurable


from .measurable_resolver import MeasurableResolver




from pydantic import BaseModel
import typing as t
import polars as pl
from loguru import logger

from uptrain.framework.signal import Signal
from uptrain.framework.checks import Check, Settings


class ValidationManager(BaseModel):
    check: t.Any
    completion_function: t.Any
    retry_logic: list[dict]
    pass_condition: t.Any
    max_retries: int = 3

    def setup(self):
        self.check.setup(Settings())

    def run(self, inputs):
        for key in inputs.keys():
            inputs[key] = [inputs[key]]
        response_is_valid = False
        num_tries = 0
        completion_function = self.completion_function
        response_history = {}
        while not response_is_valid:
            response = completion_function(inputs)
            response_history.update({'attempt_' + str(num_tries): response})
            num_tries += 1
            inputs["response"] = [response]
            data = pl.from_dict(inputs)
            data = self.check.run(data)
            response_is_valid = self.pass_condition.run(data)[0]

            if response_is_valid:
                logger.success(f"Validation check PASSED after {num_tries} attempt(s)")                
            elif num_tries < self.max_retries:
                completion_function = self.completion_function
                selected_retry_function = "default"
                for retry_condition in list(self.retry_logic):
                    if retry_condition['signal'].run(data)[0] == True:
                        completion_function = retry_condition['completion_function']
                        selected_retry_function = retry_condition['name']
                        break
                logger.warning(f"RETRYING validation check {num_tries} of {self.max_retries} with logic: {selected_retry_function}")
            else:
                response_is_valid = True
                logger.warning(f"Validation check FAILED after {self.max_retries} attempt(s)")
        if num_tries > 1:
            logger.info(f"Response History: {response_history}")

        return response



import numpy as np
import pandas as pd
from uptrain.v0.core.classes.monitors import AbstractMonitor
from uptrain.v0.ee.classes.measurables import GrammerScoreMeasurable
from uptrain.v0.constants import Monitor
from uptrain.v0.core.lib.helper_funcs import read_json

from scipy.stats import zscore


class DataIntegrity(AbstractMonitor):
    dashboard_name = "data_integrity"
    monitor_type = Monitor.DATA_INTEGRITY

    def base_init(self, fw, check):
        self.integrity_type = check["integrity_type"]
        # Threshold value for data integrity check
        self.threshold = check.get("threshold", None)
        # Threshold for when to alert on percentage of outliers (default 2%)
        self.outliers_alert_thres = check.get("outliers_alert_thres", 2)
        self.count = 0
        self.num_issues = 0
        self.has_reference_dataset = "reference_dataset" in check
        if self.has_reference_dataset:
            self.reference_dataset = check["reference_dataset"]
            self.ref_mean, self.ref_std = self.get_ref_data_stats()

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        signal_value = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        signal_value = np.reshape(np.array(signal_value), -1)

        if self.integrity_type == "non_null":
            self.has_issue = [x == None for x in signal_value]
        elif self.integrity_type == "less_than":
            self.has_issue = signal_value > self.threshold
        elif self.integrity_type == "equal_to":
            self.has_issue = signal_value == self.threshold
        elif self.integrity_type == "greater_than":
            self.has_issue = signal_value < self.threshold
        elif self.integrity_type == "minus_one":
            self.has_issue = signal_value == -1
        elif self.integrity_type == "z_score":
            if self.threshold is None:
                self.threshold = 3

            # Calculating Z-scores w.r.t. the reference dataset
            if self.has_reference_dataset:
                z_score = (signal_value - self.ref_mean) / self.ref_std
            # Calculating Z-scores w.r.t. the current dataset
            else:
                z_score = zscore(signal_value)
            self.has_issue = np.abs(z_score) > self.threshold
            outliers = signal_value[self.has_issue]
            valid_signal = signal_value[~self.has_issue]

            feat_name = self.measurable.col_name()
            plot_name = f"z_score_feature_{feat_name}"
            self.log_handler.add_histogram(
                plot_name=plot_name,
                data=valid_signal,
                dashboard_name=self.dashboard_name,
                file_name=f"valid_z_scores",
            )
            percentage_outliers = 100 * len(outliers) / len(z_score)
            if len(outliers) > 0 and percentage_outliers > self.outliers_alert_thres:
                self.log_handler.add_histogram(
                    plot_name=plot_name,
                    data=outliers,
                    dashboard_name=self.dashboard_name,
                    file_name=f"outliers",
                )
                perc = round(percentage_outliers, 1)
                self.log_handler.add_alert(
                    alert_name=f"Z-score outliers detected for {feat_name} 🚨",
                    alert=f"{len(outliers)} of {len(z_score)} samples have Z-Score > {self.threshold} ({perc}%)",
                    dashboard_name=self.dashboard_name,
                )
        elif self.integrity_type == "grammar_check":
            self.has_issue = signal_value < self.threshold
        else:
            raise NotImplementedError(
                "Data integrity check {} not implemented".format(self.integrity_type)
            )
        self.count += len(signal_value)
        self.num_issues += np.sum(np.array(self.has_issue))

        self.log_handler.add_scalars(
            self.integrity_type + "_outliers_ratio",
            {"y_outliers": self.num_issues / self.count},
            self.count,
            self.dashboard_name,
            file_name=self.measurable.col_name(),
        )

    def need_ground_truth(self):
        return False

    def base_is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        if self.integrity_type == "grammar_check":
            reasons = ["None"] * len(extra_args["id"])
            for idx in range(len(extra_args["id"])):
                if self.has_issue[idx]:
                    reasons.append(
                        "Data Integrity Issue, Type: {}".format(self.integrity_type)
                    )
            return self.has_issue, reasons
        else:
            return np.array([False] * len(extra_args["id"])), np.array(
                ["None"] * len(extra_args["id"])
            )

    def get_ref_data_stats(self):
        """
        Find the mean and std for z-score in ref_arr. The data can be numerical or categorical.
        """
        if self.reference_dataset.split(".")[-1] == "json":
            data = read_json(self.reference_dataset)
            all_inputs = np.array(
                [self.measurable.extract_val_from_training_data(x) for x in data]
            )
        elif self.reference_dataset.split(".")[-1] == "csv":
            data = pd.read_csv(self.reference_dataset).to_dict()
            for key in data:
                data[key] = list(data[key].values())
            all_inputs = np.array(self.measurable.extract_val_from_training_data(data))
        else:
            raise Exception("Reference data file type not recognized.")

        return np.mean(all_inputs), np.std(all_inputs)



from .dataset import KpsDataset, read_json, write_json, plot_all_cluster
from .model_torch import get_accuracy_torch, train_model_torch, BinaryClassification
from .pushup_signal import body_length_signal, pushup_signal
from .get_data_from_remote import get_data_from_remote
from .helper_funcs import download_dataset, process



"""
Implement checks to test if a piece of text has been taken from a source.
"""

from __future__ import annotations
import typing as t
import os
import copy 
import re


from loguru import logger
import polars as pl
from uptrain.framework import Settings

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.operators.language.llm import LLMMulticlient, Payload
# from evals.elsuite.modelgraded.classify_utils import (
#     # append_answer_prompt,
#     # get_choice,
#     # get_choice_score,
# )

import logging
import string
from typing import Any, Callable, Iterable, Optional, Union

MATCH_FNS = {
    "include": lambda x, y: float(x in y),
    "exact": lambda x, y: float(x == y),
    "endswith": lambda x, y: x.endswith(y),
    "starts_or_endswith": lambda x, y: x.startswith(y) or x.endswith(y),
}
INVALID_STR = "__invalid__"

ANSWER_PROMPTS = {
    # e.g. "Yes"
    "classify": "Answer the question by printing only a single choice from {choices} (without quotes or punctuation) corresponding to the correct answer with no other text.".strip(),
    # e.g. "Yes\n The reasons are: ..."
    "classify_cot": "First, answer by printing a single choice from {choices} (without quotes or punctuation) corresponding to the correct answer. Then, from the next line, explain your reasonings step by step.".strip(),
    # e.g. "Let's think step by step. ...\nYes"
    "cot_classify": """
You are also given scores for each choice as {choice_scores}. {choice_scores_text}

First, write out in a step by step manner your reasoning to be sure that your conclusion is correct. Avoid simply stating the correct answer at the outset.
Then print only the score from {scores_text} associated to the correct answer in a separate line. Finally repeat the same score on a new line.
Reasoning:""",
    "tot_classify": "",
    "tot_score": "",
}

def get_choice_score(
    choice: str,
    choice_strings: Iterable[str],
    choice_scores: Optional[Union[dict[str, float], str]] = None,
) -> Optional[float]:
    if choice_scores is None:
        return None
    if choice_scores == "from_strings":
        choice_scores = {c: float(c) for c in choice_strings}
    # assumption: each INVALID_STR contributes the lowest score
    if choice == INVALID_STR:
        return min(choice_scores.values())
    return choice_scores[choice]

def choice_to_str(choice_strings: Iterable[str]) -> str:
    """Return a string of choices, e.g. '"Yes" or "No" or "Maybe"'."""
    return " or ".join(f'"{choice}"' for choice in choice_strings)

def append_answer_prompt(
    prompt: list,
    eval_type: str,
    append_type: str = "as_content",
    answer_prompt: Optional[list] = None,
    choice_strings: Optional[Iterable[str]] = None,
) -> list:
    """Append answer prompt to prompt."""
    answer_prompt = answer_prompt or ANSWER_PROMPTS[eval_type] #.format(**{"choice_strings": ','.join(choice_strings)})
    answer_prompt = format_prompt(answer_prompt, choices=choice_to_str(choice_strings))
    if append_type == "as_content":
        assert isinstance(answer_prompt, str), f"prompt must be str, not {type(answer_prompt)}"
        prompt[-1]["content"] += "\n\n" + answer_prompt
    elif append_type == "as_message":
        assert is_chat_prompt(answer_prompt), f"prompt must be chat prompt, not {answer_prompt}"
        prompt += answer_prompt
    else:
        raise ValueError(f"append_type must be 'as_content' or 'as_message', not {append_type}")
    return prompt


def is_chat_prompt(prompt) -> bool:
    return isinstance(prompt, list) and all(isinstance(msg, dict) for msg in prompt)

def chat_prompt_to_text_prompt(prompt: list, for_completion: bool = True) -> str:
    """
    Render a chat prompt as a text prompt. User and assistant messages are separated by newlines
    and prefixed with "User: " and "Assistant: ", respectively, unless there is only one message.
    System messages have no prefix.
    """
    assert is_chat_prompt(prompt), f"Expected a chat prompt, got {prompt}"
    chat_to_prefixes = {
        # roles
        "system": "",
        # names
        "example_user": "User: ",
        "example_assistant": "Assistant: ",
    }

    # For a single message, be it system, user, or assistant, just return the message
    if len(prompt) == 1:
        return prompt[0]["content"]

    text = ""
    for msg in prompt:
        role = msg["name"] if "name" in msg else msg["role"]
        prefix = chat_to_prefixes.get(role, role.capitalize() + ": ")
        content = msg["content"]
        text += f"{prefix}{content}\n"
    if for_completion:
        text += "Assistant: "
    return text.lstrip()

def format_necessary(template: str, allow_missing: bool = False, **kwargs: dict[str, str]) -> str:
    """Format a template string with only necessary kwargs."""
    keys = [k[1] for k in string.Formatter().parse(template) if k[1]]
    if allow_missing:
        assert (
            len([k for k in keys if k in kwargs]) > 0
        ), f"Required: {keys}, got: {sorted(kwargs)}, no inputs are used.\nTemplate:\n{template}"
        cur_keys = {k: kwargs.get(k, "{" + k + "}") for k in keys}
    else:
        assert all(
            k in kwargs for k in keys
        ), f"Required: {keys}, got: {sorted(kwargs)}.\nTemplate:\n{template}"
        cur_keys = {k: kwargs[k] for k in keys}
    return template.format(**cur_keys)


def format_prompt(
    prompt: list, allow_missing: bool = False, **kwargs: dict[str, str]
) -> list:
    """Format a prompt with only necessary kwargs."""
    # if any input kwargs is chat prompt, convert to text prompt
    kwargs = {
        k: chat_prompt_to_text_prompt(v, for_completion=False) if is_chat_prompt(v) else v
        for k, v in kwargs.items()
    }
    if is_chat_prompt(prompt):
        new_prompt = []
        for msg in prompt:
            formatted_msg = copy.copy(msg)
            if "content" in formatted_msg:
                formatted_msg["content"] = format_necessary(
                    formatted_msg["content"], allow_missing=allow_missing, **kwargs
                )
            new_prompt.append(formatted_msg)
        prompt = new_prompt
    else:
        # Prompt is a string
        prompt = format_necessary(prompt, allow_missing=allow_missing, **kwargs)
    return prompt


@register_op
class OpenAIGradeScore(ColumnOp):
    """
    Operator to calculate the grade score of text completions using OpenAI models.

    Attributes:
        col_in_input (str): The name of the input column containing the prompts.
        col_in_completion (str): The name of the input column containing the completions.
        eval_name (str): The name of the OpenAI evaluation to use.
        col_out (str): The name of the output column containing the grade scores.

    Returns:
        dict: A dictionary containing the calculated grade scores.

    """

    col_in_input: str = "prompt"
    col_in_completion: str = "response"
    col_out: str = "openai_grade_score"
    eval_name: str

    def setup(self, settings: Settings):
        self._settings = settings
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        samples = data.select(
            [
                pl.col(self.col_in_input).alias("input"),
                pl.col(self.col_in_completion).alias("completion"),
            ]
        )
        from uptrain.operators.language.openai_evals import OpenaiEval
        grading_op = OpenaiEval(
            bundle_path="",
            completion_name="gpt-3.5-turbo",
            eval_name=self.eval_name,
        )

        grading_op.setup(settings=self._settings)
        oaieval_res = grading_op.run(samples)
        assert (
            "extra" in oaieval_res
            and "metrics" in oaieval_res["extra"]
            and "score" in oaieval_res["extra"]["metrics"]
        )

        results = pl.Series(oaieval_res["extra"]["metrics"]["score"])
        return {"output": data.with_columns([results.alias(self.col_out)])}


@register_op
class ModelGradeScore(ColumnOp):
    """
    Operator to calculate the grade score of text completions using a custom prompt
    for grading. It is a wrapper using the same utilities from the OpenAI evals library,
    replacing just the completion call.

    Attributes:
        grading_prompt_template (str): Template for the grading prompt.
        eval_type (Literal["cot_classify", "classify", "classify_cot"]): The type of evaluation for grading ("cot_classify" by default).
        choice_strings (list[str]): The list of choice strings for grading.
        choice_scores (dict[str, float]): The dictionary mapping choice strings to scores.
        context_vars (dict[str, str]): A dictionary mapping context variable names to corresponding
            columns in the input dataset.
    """

    grading_prompt_template: str
    eval_type: t.Literal["cot_classify", "classify", "classify_cot", 'tot_classify', 'tot_score'] = "cot_classify"
    choice_strings: list[str]
    choice_scores: dict[str, float]  #t.Union[dict[str, float], dict[str, list[float]]]
    context_vars: dict[str, str]
    col_out: t.Union[str, list[str]] = "model_grade_score"

    def setup(self, settings: Settings):
        self._api_client = LLMMulticlient(settings=settings)
        self._settings = settings
        self.model = settings.model.replace("azure/", "")
        if not (self.eval_type in ["cot_classify", "tot_classify", 'tot_score']):
            raise Exception("Only eval_type: cot_classify and tot_classify is supported for model grading check")
        for choice, score in self.choice_scores.items():
            score = format(score, ".3f")
            if score[-1] == "0":
                score = score[0:-1]
                if score[-1] == "0":
                    score = score[0:-1]
            self.choice_scores[choice] = score
        choice_scores = "(" + str(self.choice_scores)[1:-1] + ")"
        choice_scores_text = ""
        for choice, score in self.choice_scores.items():
            choice_scores_text += f"If selected choice is {choice}, score should be {score}. "
        scores_text = "(" + ", ".join(list(self.choice_scores.values())) + ")"
        answer_prompt = ANSWER_PROMPTS[self.eval_type].format(choice_scores=choice_scores, choice_scores_text=choice_scores_text, scores_text=scores_text)
        self.grading_prompt_template += answer_prompt
        return self

    def _make_payload(self, id: t.Any, messages: list[dict]) -> Payload:
        if self._settings.seed is None:
            return Payload(
                data={"model": self.model, "messages": messages, "temperature": 0.2},
                metadata={"index": id},
            )
        else:
            return Payload(
                data={"model": self.model, "messages": messages, "temperature": 0.2, "seed": self._settings.seed},
                metadata={"index": id},
            )

    def get_choice_via_llm(self, text: str, grading_prompt_template: str) -> str:
        """Queries LLM to get score from the text"""

        prompt = f"""
        Extract the score from the given text. The available choices and associated scores is present in the context.

        Context: {grading_prompt_template}
        Text: {text}

        Score:
        """

        payload = self._make_payload(0, [{"role": "user", "content": prompt}])
        output_payload = self._api_client.fetch_responses([payload])[0]

        try:
            score = output_payload.response.choices[0].message.content
            float(score)
            return score
        except:
            return str(0.0)


    def get_choice(
        self, text: str, eval_type: str, match_fn: Union[str, Callable], choice_strings: Iterable[str], choice_scores: dict = {}
    ) -> str:
        """Clean the answer string to a choice string to one of choice_strings. Return '__invalid__.' if no match."""
        if eval_type == "tot_score":
            score = ''
            if len(score) == 0:
                scores_matches = re.findall(r"(\[Score\]\: [1-5]|Score: [1-5]|Score is [1-5])", text)
                if len(scores_matches)!=0:
                    score = str(scores_matches[0].split(' ')[-1])

            if len(score) == 0:
                scores_matches = re.findall(r"(\[score\]\: [1-5]|score: [1-5]|score is [1-5])", text)
                if len(scores_matches)!=0:
                    score = str(scores_matches[0].split(' (')[-1])

            if len(score) == 1:
                return float(score)

            logging.warn(f"Choices {choice_strings} not parsable for {eval_type}: {text}")
            return -5
        elif eval_type == "tot_classify":
            score = ''
            if len(score) == 0:
                scores_matches = re.findall(r"(\[Choice\]\: [a-zA-Z]|Choice: [a-zA-Z]|choice is [a-zA-Z])", text)
                if len(scores_matches)!=0:
                    score = str(scores_matches[0].split(' ')[-1])

            if len(score) == 0:
                scores_matches = re.findall(r"(\[Choice\]\: \([a-zA-Z]|Choice: \([a-zA-Z]|choice is \([a-zA-Z])", text)
                if len(scores_matches)!=0:
                    score = str(scores_matches[0].split(' (')[-1])

            if len(score) == 0:
                scores_matches = re.findall(r"(\[Argument\]\: \([a-zA-Z]|Argument: \([a-zA-Z]|argument is \([a-zA-Z])", text)
                if len(scores_matches)!=0:
                    score = str(scores_matches[0].split(' (')[-1])

            if len(score) == 1:
                if score.upper() in choice_scores:
                    return choice_scores[score.upper()]
                elif score.lower() in choice_scores:
                    return choice_scores[score.lower()]

            logging.warn(f"Choices {choice_strings} not parsable for {eval_type}: {text}")
            return -1
        else:
            is_fn_extract_score = False
            if match_fn == 'extract_score':
                is_fn_extract_score = True
            else:
                if isinstance(match_fn, str):
                    match_fn = MATCH_FNS[match_fn]
            lines = text.strip().split("\n")
            if eval_type.startswith("cot_classify"):
                lines = lines[::-1]  # reverse lines
            for line in lines:
                line = line.strip()
                if not line:
                    continue
                if is_fn_extract_score:
                    if "." in line:
                        part_before_decimal = line.split(".")[0][::-1]
                        prev_char = ""
                        for char in part_before_decimal:
                            new_char = char + prev_char
                            try:
                                float(new_char)
                            except:
                                break
                            prev_char = new_char
                        part_before_decimal = prev_char

                        part_after_decimal = line.split(".")[1]
                        prev_char = ""
                        for char in part_after_decimal:
                            new_char = prev_char + char
                            try:
                                float(new_char)
                            except:
                                break
                            prev_char = new_char
                        part_after_decimal = prev_char
                        choice = part_before_decimal + "." + part_after_decimal
                        try:
                            float(choice)
                            if float(choice) > 1.0 or float(choice) < 0.0:
                                return self.get_choice_via_llm(text, self.grading_prompt_template)
                            return str(choice)
                        except:
                            return self.get_choice_via_llm(text, self.grading_prompt_template)
                    else:
                        return self.get_choice_via_llm(text, self.grading_prompt_template)
                else:
                    line = "".join(c for c in line if c not in string.punctuation)
                    if not line:
                        continue
                    for choice in choice_strings:
                        if match_fn(line, choice):
                            return choice
            logging.warn(f"Choices {choice_strings} not parsable for {eval_type}: {text}")
            return INVALID_STR

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        prompts = []
        for row in data.rows(named=True):
            subs = {k: row[v] for k, v in self.context_vars.items()}
            # fill in context variables in the prompt template
            _prompt = self.grading_prompt_template.format(**subs)
            # following the `evals` code to create the grading instruction
            #  https://github.com/openai/evals/blob/main/evals/elsuite/modelgraded/classify_utils.py
            _prompt_chat = [{"role": "user", "content": _prompt}]
            # _prompt_chat = append_answer_prompt(
            #     prompt=[{"role": "user", "content": _prompt}],
            #     eval_type=self.eval_type,
            #     choice_strings=self.choice_strings,
            # )
            prompts.append(_prompt_chat)

        input_payloads = [
            self._make_payload(idx, prompt_msgs)
            for idx, prompt_msgs in enumerate(prompts)
        ]
        output_payloads = self._api_client.fetch_responses(input_payloads)

        results = []
        for res in output_payloads:
            idx = res.metadata["index"]
            if res.error is not None:
                logger.error(
                    f"Error when processing payload at index {idx}: {res.error}"
                )
                results.append((idx, None, None))
            else:
                try:
                    resp_text = res.response.choices[0].message.content
                    choice = self.get_choice(
                        text=resp_text,
                        eval_type=self.eval_type,
                        match_fn="extract_score",
                        choice_strings=self.choice_strings,
                        choice_scores = self.choice_scores
                    )
                    score = float(choice)
                    results.append((idx, score, resp_text))
                except Exception as e:
                    logger.error(
                        f"Error when processing payload at index {idx}, not API error: {e}"
                    )
                    results.append((idx, None, None))

        results = sorted(results, key=lambda x: x[0])
        if isinstance(self.col_out, list):
            result_scores = [
                pl.Series(
                    [val[idx] if val is not None else None for _, val, _ in results]
                ).alias(self.col_out[idx])
                for idx in range(len(self.col_out))
            ]
            result_scores.extend(
                [
                    pl.Series([explanation for _, _, explanation in results]).alias(
                        self.col_out[idx] + "_explanation"
                    )
                    for idx in range(len(self.col_out))
                ]
            )
        else:
            result_scores = [
                pl.Series([val for _, val, _ in results]).alias(self.col_out),
                pl.Series([explanation for _, _, explanation in results]).alias(
                    self.col_out + "_explanation"
                ),
            ]
        return {"output": data.with_columns(result_scores)}



from typing import Optional, Union, TYPE_CHECKING

try:
    import umap
except:
    umap = None
import numpy as np
from sklearn.manifold import TSNE

from uptrain.v0.core.lib.helper_funcs import cluster_and_plot_data
from uptrain.v0.core.classes.visuals import AbstractVisual, ClusteringResolver
from uptrain.v0.constants import Visual, Statistic, MeasurableType, ClusteringAlgorithm
from uptrain.v0.core.classes.measurables import MeasurableResolver
from uptrain.v0.core.lib.helper_funcs import read_json, dependency_required

if TYPE_CHECKING:
    from uptrain.v0.core.classes.logging.new_log_handler import (
        LogHandler as NewLogHandler,
        LogWriter,
    )
    from uptrain.v0.core.classes.logging.log_handler import LogHandler


@dependency_required(umap, "umap-learn")
class DimensionalityReduction(AbstractVisual):
    log_handler: Union["LogHandler", "NewLogHandler"]
    log_writer: Optional["LogWriter"]

    def base_init(self, fw, check):
        self.visual_type = check["type"]
        self.dashboard_name = check.get("dashboard_name", "visual")
        if self.visual_type == Visual.UMAP:
            self.umap_init(check)
        elif self.visual_type == Visual.TSNE:
            self.tsne_init(check)
        else:
            raise Exception("Dimensionality reduction type undefined.")
        self.framework = fw
        self.clustering_algorithm = ClusteringResolver(
            check.get("clustering_algorithm", ClusteringAlgorithm.HDBSCAN)
        ).resolve(check.get("clustering_args", {}))
        self.label_measurables = [
            MeasurableResolver(x).resolve(fw) for x in check.get("label_args", [])
        ]
        label_names = [x.col_name() for x in self.label_measurables]
        self.labels = {"clusters": []}
        for label_name in label_names:
            self.labels[label_name] = []
        self.hover_measurables = [
            MeasurableResolver(x).resolve(fw) for x in check.get("hover_args", [])
        ]
        self.hover_names = [x.col_name() for x in self.hover_measurables]
        if "id" not in self.hover_names:
            id_args = {"type": MeasurableType.INPUT_FEATURE, "feature_name": "id"}
            self.hover_measurables.append(MeasurableResolver(id_args).resolve(fw))
            self.hover_names.append("id")
        self.count_checkpoints = check.get("count_checkpoints", ["all"])
        self.dim = check.get("dim", "2D")

        self.total_count = 0
        self.prev_calc_at = 0
        self.update_freq = check.get("update_freq", 10000)
        self.vals = []
        self.hover_texts = []
        self.feature_dictn = {}

        # get handles to the log writer object
        if hasattr(self.log_handler, "make_logger"):
            self.log_writer = self.log_handler.make_logger(
                self.dashboard_name, self.visual_type, fmt="json"
            )
        else:
            self.log_writer = None

        self.initial_dataset = check.get("initial_dataset", None)
        if self.initial_dataset is not None:
            data = read_json(self.initial_dataset)
            self.vals.extend(
                [self.measurable.extract_val_from_training_data(x) for x in data]
            )

            for label_measurable in self.label_measurables:
                label_data = [
                    label_measurable.extract_val_from_training_data(x) for x in data
                ]
                self.labels[label_measurable.col_name()] = label_data
            if len(self.hover_measurables):
                offset = 0
                for data_point in data:
                    this_hovers = [
                        x.extract_val_from_training_data(data_point)
                        for x in self.hover_measurables
                    ]
                    self.hover_texts.append(dict(zip(self.hover_names, this_hovers)))
                    # self.hover_texts[offset].update({"Training Density": int(points_density[offset])})
                    self.hover_texts[offset].update({"idx": offset})
                    offset += 1
            if len(self.feature_measurables):
                all_features = [
                    list(
                        np.reshape(
                            [
                                feature_measurable.extract_val_from_training_data(x)
                                for x in data
                            ],
                            -1,
                        )
                    )
                    for feature_measurable in self.feature_measurables
                ]
                this_dict = dict(
                    zip(["feature_" + x for x in self.feature_names], all_features)
                )
                for key in this_dict.keys():
                    if key in self.feature_dictn:
                        self.feature_dictn[key].extend(this_dict[key])
                    else:
                        self.feature_dictn.update({key: list(this_dict[key])})

    def umap_init(self, check):
        self.min_dist = check.get("min_dist", 0.01)
        self.n_neighbors = check.get("n_neighbors", 20)
        self.metric_umap = check.get("metric", "euclidean")

    def tsne_init(self, check):
        self.perplexity = check.get("perplexity", 30.0)
        self.early_exaggeration = check.get("early_exaggeration", 12.0)
        self.learning_rate = check.get("learning_rate", "auto")
        self.n_iter = check.get("n_iter", 1000)
        self.n_iter_without_progress = check.get("n_iter_without_progress", 300)
        self.min_grad_norm = check.get("min_grad_norm", 1e-7)
        self.metric_tsne = check.get("metric", "euclidean")
        self.metric_params = check.get("metric_params", None)
        self.init = check.get("init", "pca")
        self.verbose = check.get("verbose", 0)
        self.random_state = check.get("random_state", None)
        self.method = check.get("method", "barnes_hut")
        self.angle = check.get("angle", 0.5)
        self.n_jobs = check.get("n_jobs", None)

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        if self.measurable is not None:
            vals = self.measurable.compute_and_log(
                inputs, outputs, gts=gts, extra=extra_args
            )
            self.vals.extend(vals)

        for label_measurable in self.label_measurables:
            label_data = label_measurable.compute_and_log(
                inputs, outputs, gts=gts, extra=extra_args
            )
            self.labels[label_measurable.col_name()].extend(label_data)

        if len(self.feature_measurables):
            all_features = [
                x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
                for x in self.feature_measurables
            ]
            this_dict = dict(
                zip(["feature_" + x for x in self.feature_names], all_features)
            )
            for key in this_dict.keys():
                if key in self.feature_dictn:
                    self.feature_dictn[key].extend(this_dict[key])
                else:
                    self.feature_dictn.update({key: list(this_dict[key])})

        if len(self.hover_measurables):
            hover_values = [
                x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
                for x in self.hover_measurables
            ]
            hover_table = np.array(hover_values).transpose()
            self.hover_texts.extend(
                [dict(zip(self.hover_names, values)) for values in hover_table]
            )

        self.total_count += len(extra_args["id"])
        if not ((self.total_count - self.prev_calc_at) >= self.update_freq):
            return

        self.prev_calc_at = self.total_count
        models = dict(
            zip(
                ["model_" + x for x in self.model_names],
                [
                    self.allowed_model_values[jdx][0]
                    for jdx in range(len(self.model_names))
                ],
            )
        )

        for count in self.count_checkpoints:
            emb_list, label_dict, hover_texts = self.get_high_dim_data(count)
            emb_list = np.array(emb_list)

            if emb_list.shape[0] > 10:
                clusters = []
                umap_list, clusters = self.get_embs_and_clusters(emb_list)
                this_data = {"umap": umap_list}
                label_dict["clusters"].extend(clusters)
                this_data.update({"labels": label_dict})
                if len(hover_texts) > 0:
                    this_data.update({"hover_texts": hover_texts})

                if self.log_writer is not None:
                    self.log_writer.log(
                        {
                            "count": str(count),
                            **this_data,
                            **models,
                            **self.feature_dictn,
                        }
                    )
                else:
                    self.log_handler.add_histogram(
                        self.visual_type,
                        this_data,
                        self.dashboard_name,
                        models=models,
                        features=self.feature_dictn,
                        file_name=self.dashboard_name
                        + "_"
                        + str(count)
                        + "_"
                        + "_".join(list(models.values())),
                    )

    def get_high_dim_data(self, count):
        if self.measurable is None:
            distribution_anomaly = list(
                filter(
                    lambda x: x.statistic_type == Statistic.DISTRIBUTION_STATS,
                    self.framework.check_manager.statistics_to_check,
                )
            )[0]
            data_dict = distribution_anomaly.get_feats_for_clustering(
                count, self.allowed_model_values
            )
            chosen_label_key = None
            chosen_hover_key = None
            if len(data_dict):
                temp_val = list(data_dict.values())[0]
                if len(temp_val):
                    temp_keys = list(temp_val.keys())
                    temp_keys = list(filter(lambda x: "visual_label_" in x, temp_keys))
                    if len(temp_keys) > 1:
                        print(
                            "Have multiple labels - "
                            + str(temp_keys)
                            + " .Using "
                            + temp_keys[0]
                            + " for labeling."
                        )
                    if len(temp_keys) > 0:
                        chosen_label_key = temp_keys[0]
                if len(temp_val):
                    temp_keys = list(temp_val.keys())
                    temp_keys = list(
                        filter(lambda x: "visual_hover_text_" in x, temp_keys)
                    )
                    if len(temp_keys) > 1:
                        print(
                            "Have multiple hover texts - "
                            + str(temp_keys)
                            + " .Using "
                            + temp_keys[0]
                            + " for hovering."
                        )
                    if len(temp_keys) > 0:
                        chosen_hover_key = temp_keys[0]
            vals = np.array([data_dict[x]["val"] for x in data_dict])
            if vals.shape[0] > 0:
                vals = np.squeeze(vals, axis=1)
            labels = []
            if chosen_label_key is not None:
                labels = [data_dict[x][chosen_label_key] for x in data_dict]

            hover_texts = []
            if chosen_hover_key is not None:
                hover_texts = [
                    {chosen_hover_key: data_dict[x][chosen_hover_key]}
                    for x in data_dict
                ]
            return vals, labels, hover_texts
        else:
            return self.vals, self.labels, self.hover_texts

    def get_embs_and_clusters(self, emb_list):
        if self.dim == "2D":
            n_components = 2
        else:
            n_components = 3

        if self.visual_type == Visual.UMAP:
            compressed_embeddings = umap.UMAP(
                n_neighbors=self.n_neighbors,
                n_components=n_components,
                min_dist=self.min_dist,
                metric=self.metric_umap,
            ).fit_transform(emb_list)
        elif self.visual_type == Visual.TSNE:
            compressed_embeddings = TSNE(
                n_components=n_components,
                perplexity=self.perplexity,
                early_exaggeration=self.early_exaggeration,
                learning_rate=self.learning_rate,
                n_iter=self.n_iter,
                n_iter_without_progress=self.n_iter_without_progress,
                min_grad_norm=self.min_grad_norm,
                metric=self.metric_tsne,
                init=self.init,
                verbose=self.verbose,
                random_state=self.random_state,
                method=self.method,
                angle=self.angle,
                n_jobs=self.n_jobs,
            ).fit_transform(emb_list)

        clustering = self.clustering_algorithm.fit(compressed_embeddings)
        labels = np.squeeze(np.array(clustering.labels_))
        return compressed_embeddings, labels



import numpy as np

from typing import List, Union

from uptrain.v0.core.classes.distances import AbstractDistance


class L2Distance(AbstractDistance):
    """Class that computes L2 distance between base and reference vectors."""

    def compute_distance(
        self, base: Union[List, np.ndarray], reference: Union[List, np.ndarray]
    ) -> np.ndarray:
        """Compute the L2 distance between base and reference vectors.

        Parameters
        ----------
        base
            It is the base vector for L2 distance computation.
        reference
            It is the reference vector for L2 distance computation.

        Returns
        -------
        np.ndarray
            The L2 distance between base and reference.
        """

        base = np.array(base)
        reference = np.array(reference)

        self.check_compatibility(base, reference)

        return np.sum(
            (base - reference) * (base - reference),
            axis=tuple(range(1, len(base.shape))),
        )



from typing import Any

from uptrain.v0.core.classes.measurables import (
    Measurable,
    InputFeatureMeasurable,
    OutputFeatureMeasurable,
)


class FeatureMeasurable(Measurable):
    """Class that returns a feature as a result of computation."""

    def __init__(self, framework, feature_name, dictn_type: str) -> None:
        """Constructor for FeatureMeasurable class.
        
        Parameters
        ----------
        feature_name
            Name of the feature that is to be measured/evaluated
        dictn_type
            Must be "inputs" or "outputs" depending on which feature to
            measure/evaluate
        """
        super().__init__(framework)
        self.feature_name = feature_name
        self.dictn_type = dictn_type
        if self.dictn_type == "inputs":
            self.helper = InputFeatureMeasurable(framework, feature_name)
        elif self.dictn_type == "outputs":
            self.helper = OutputFeatureMeasurable(framework, feature_name)
        else:
            raise Exception(
                "Helper Measurable not defined for dictionary type %s" % self.dictn_type
            )

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> Any:
        return self.helper._compute(
            inputs=inputs, outputs=outputs, gts=gts, extra=extra
        )

    def col_name(self) -> str:
        return self.helper.col_name()

    # TODO: Decommission and find a generic way
    def extract_val_from_training_data(self, x):
        return self.helper.extract_val_from_training_data(x)



import time
import uptrain.v0 as v0
import matplotlib
import pandas as pd

from sklearn import model_selection
from lightgbm import LGBMRegressor
from test_helpers import download_dataset, process

matplotlib.use("TkAgg")

data_file = "trip_duration_dataset.csv"
download_dataset(data_file)
df_train = pd.read_csv(data_file)

# Divide the data into input data to the model and targets
y = df_train["trip_duration"]
X = df_train.drop(["trip_duration"], axis=1)

# Divide data into train-test splits
X_train, X_test, y_train, y_test = model_selection.train_test_split(X, y, test_size=0.1)

# Get ids for data to locate them with their ids in the UpTrain dashboard
test_ids = list(X_test["id"])

# Clean up data
X_train = process(X_train)
X_test = process(X_test)

# Train the model
model = LGBMRegressor(n_estimators=500)
model.fit(X_train, y_train)

# Create a shap visualization
shap_visual = {
    "type": v0.Visual.SHAP,
    "model": model,
    "shap_num_points": 10000,
}

cfg = {
    "checks": [shap_visual],
    "logging_args": {"st_logging": True},
}

framework = v0.Framework(cfg_dict=cfg)

batch_size = 2000
cols = list(X_test.columns)
for idx in range(int(len(X_test) / batch_size)):
    indices = range(idx * batch_size, (idx + 1) * batch_size)
    this_elems = X_test.iloc[idx * batch_size : (idx + 1) * batch_size, :]
    this_preds = abs(model.predict(this_elems))

    # Add ids to corresponding data points to preserve them in dashboard
    this_elems = this_elems.assign(
        id=list(test_ids[idx * batch_size : (idx + 1) * batch_size])
    )

    # Log input and outputs to the UpTrain framework
    ids = framework.log(inputs=this_elems, outputs=this_preds)

    # Attach ground truth
    ground_truth = list(y_test[idx * batch_size : (idx + 1) * batch_size])
    framework.log(identifiers=ids, gts=ground_truth)

    time.sleep(0.01)



"""
Implement checks to test similarity between two pieces of text.

This module provides the `CosineSimilarity` class, which calculates the cosine similarity between two vectors representing text. The vectors can be columns in a DataFrame. The cosine similarity measures the similarity between two vectors based on the cosine of the angle between them.

"""

from __future__ import annotations
import typing as t

import numpy as np
from loguru import logger
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *


@register_op
class CosineSimilarity(ColumnOp):
    """
    Column operation to calculate the cosine similarity between two vectors representing text.

    Attributes:
        col_in_vector_1 (str): The name of the column containing the first vector.
        col_in_vector_2 (str): The name of the column containing the second vector.
        col_out (str): The name of the output column containing the cosine similarity scores.

    Returns:
        dict: A dictionary containing the cosine similarity scores.

    Example:
        ```
        import polars as pl
        import numpy as np
        from uptrain.operators import CosineSimilarity

        # Create a DataFrame
        df = pl.DataFrame({
            "vector_1": [np.array([0.1, 0.2, 0.3]), np.array([0.4, 0.5, 0.6])],
            "vector_2": [np.array([0.7, 0.8, 0.9]), np.array([0.2, 0.3, 0.4])]
        })

        # Create an instance of the CosineSimilarity class
        similarity_op = CosineSimilarity(col_in_vector_1="vector_1", col_in_vector_2="vector_2")

        # Calculate the cosine similarity between the two vectors
        result = similarity_op.run(df)
        similarity_scores = result["output"]

        # Print the similarity scores
        print(similarity_scores)
        ```

    Output:
        ```
        shape: (2,)
        Series: '_col_0' [f64]
        [
                1.861259
                0.288437
        ]
        ```

    """

    col_in_vector_1: str
    col_in_vector_2: str
    col_out: str = "cosine_similarity"

    def setup(self, settings: Settings):
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        vector_1 = data.get_column(self.col_in_vector_1)
        vector_2 = data.get_column(self.col_in_vector_2)

        results = []
        for i in range(len(vector_1)):
            v1 = np.array(vector_1[i])
            v2 = np.array(vector_2[i])
            similarity_score = np.dot(v1, v2) / np.linalg.norm(v1) * np.linalg.norm(v2)
            results.append(similarity_score)

        return {"output": data.with_columns(pl.Series(results).alias(self.col_out))}



import numpy as np
import string
import uptrain.v0 as v0


def test_concept_drift_ddm():
    """Test concept drift detection using DDM algorithm.

    We generate a synthetic dataset consisting of 1000 strings, each of length
    between 10 to 20 characters, containing lowercase alphabets, digits, and
    spaces. We also generate a binary sentiment label for each string. A label
    of 0 means that the sentiment is negative, and a label of 1 means that the
    sentiment is positive.

    Note that you can use a real dataset as well. However, for the sake of
    simplicity, we use a synthetic dataset to demonstrate how DDM works.

    Assume that we have a model that has been trained to predict the sentiment
    of a sentence. It performs well for the most part but occasionally makes
    incorrect predictions. However, for some reason, it starts to perform poorly
    on some newer data. How do we detect this?

    We use the DDM algorithm to detect the concept drift. The drift will arise
    somewhere around the 800th sentence.

    To introduce concept drift to our synthetic dataset, we randomly flip the
    predictions of 10% of the strings. This simulates a scenario where the model
    makes incorrect predictions on some data. We also deliberately make the last
    20% of predictions random. This simulates a scenario where the model encounters
    new or unseen data, causing the predictions to become inaccurate.

    In this test, we use the UpTrain framework to log the inputs, outputs, and ground
    truths. We set up the framework to use the DDM algorithm to monitor the error rate
    of the model. The error rate is calculated as the binary accuracy between the
    predictions and the ground truths. We set the threshold for warning to 2.0, and
    drift to 3.0. This means that if the error rate crosses 2.0, the model has
    begun to perform poorly on some data. If the error rate crosses 3.0, a concept
    drift alert will be raised. This means that the model is performing really poorly
    and is inaccurate on some recent data that it has been exposed to.

    The test passes if the DDM algorithm successfully detects the concept drift and
    raises an alert around the point where the predictions start becoming inaccurate.
    """

    random_state = np.random.RandomState(seed=1337)
    alphabet = np.array(list(map(ord, string.ascii_lowercase + string.digits + " ")))

    def random_sentence(length):
        """Generate a random sentence of given length"""
        return "".join(list(map(chr, random_state.choice(alphabet, size=length))))

    n = 1000
    inputs = np.array([random_sentence(random_state.randint(10, 20)) for _ in range(n)])
    sentiments = np.array([random_state.choice([0, 1]) for _ in range(n)])
    predictions = sentiments.copy()

    # Make 10% predictions incorrect randomly, i.e., if a sentiment is correctly classified,
    # make it incorrectly classified and vice versa
    incorrect_predictions = random_state.choice(n, size=int(n * 0.1), replace=False)
    predictions[incorrect_predictions] = 1 - predictions[incorrect_predictions]

    # Deliberately make the last 20% of predictions random
    random_predictions = np.arange(n - int(n * 0.2), n)
    predictions[random_predictions] = random_state.choice(
        [0, 1], size=len(random_predictions)
    )

    # Create a configuration for the framework
    cfg = {
        # Define checks to be performed
        "checks": [
            # Check: Concept drift detection using DDM algorithm
            # with accuracy as the measurable for the feature "data"
            {
                "type": v0.Monitor.CONCEPT_DRIFT,
                "algorithm": v0.DataDriftAlgo.DDM,
                "measurable_args": {"type": v0.MeasurableType.ACCURACY},
                "warm_start": 200,
                "warn_threshold": 2.0,
                "alarm_threshold": 3.0,
            }
        ],
        # Specify where the logging data should be stored
        "retraining_folder": "uptraining_smart_data_concept_drift_ddm",
        # Specify logging arguments
        # "st_logging" should be True if we want streamlit logging, False otherwise
        "logging_args": {"st_logging": True},
    }

    framework = v0.Framework(cfg)
    batch_size = 64

    # Feed the data to the framework
    for i in range(n // batch_size):
        input = {"data": inputs[i * batch_size : (i + 1) * batch_size]}
        output = predictions[i * batch_size : (i + 1) * batch_size]
        gts = sentiments[i * batch_size : (i + 1) * batch_size]
        ids = framework.log(inputs=input, outputs=output)
        framework.log(identifiers=ids, gts=gts)



__all__ = [
    # base
    "Operator",
    "ColumnOp",
    "TransformOp",
    "register_op",
    "register_custom_op",
    "get_output_col_name_at",
    "deserialize_operator",
    # drift
    "ParamsDDM",
    "ParamsADWIN",
    "ConceptDrift",
    # embs
    "Distribution",
    "UMAP",
    # clustering
    "Clustering",
    # table
    "ColumnExpand",
    "ColumnComparison",
    "ColumnReduce",
    # metrics
    "Accuracy",
    # similarity
    "CosineSimilarity",
    # vis
    "BarChart",
    "Histogram",
    "LineChart",
    "ScatterPlot",
    "MultiPlot",
    "Table",
    "CustomPlotlyChart",
    "Scatter3DPlot",
    # language - also include all the subimports
    "language",
    "GrammarScore",
    "OpenaiEval",
    "PromptEval",
    "Embedding",
    "RougeScore",
    "BLEUScore",
    "METEORScore",
    "DocsLinkVersion",
    "TextLength",
    "WordCount",
    "TextComparison",
    "KeywordDetector",
    "OpenAIGradeScore",
    "ModelGradeScore",
    "PromptGenerator",
    "TopicGenerator",
    "TextCompletion",
    "OutputParser",
    "ResponseFactualScore",
    "ContextRelevance",
    "ResponseRelevance",
    "ResponseCompleteness",
    "ResponseCompletenessWrtContext",
    "ResponseConsistency",
    "ResponseConciseness",
    "LanguageCritique",
    "ToneCritique",
    "GuidelineAdherenceScore",
    "PromptInjectionScore",
    "ConversationSatisfactionScore",
    "ResponseMatchingScore",
    "ValidResponseScore",
    "TopicAssignmentviaCluster",
    # io - also include all the subimports
    "io",
    "ExcelReader",
    "CsvReader",
    "JsonReader",
    "JsonWriter",
    "DeltaReader",
    "DeltaWriter",
    "BigQueryReader",
    "MongoDBReader",
    "DuckDBReader",
    "BigQueryWriter",
    # code
    "code",
    "ParseCreateStatements",
    "ParseSQL",
    "ValidateTables",
    "ExecuteAndCompareSQL",
]

from .base import (
    Operator,
    ColumnOp,
    TransformOp,
    register_op,
    register_custom_op,
    get_output_col_name_at,
    deserialize_operator,
)
from .drift import ConceptDrift, ParamsADWIN, ParamsDDM
from .embs import Distribution, UMAP
from .clustering import Clustering
from .table import Table, ColumnExpand, ColumnComparison, ColumnReduce
from .metrics import Accuracy
from .similarity import CosineSimilarity
from .chart import (
    BarChart,
    LineChart,
    ScatterPlot,
    Histogram,
    MultiPlot,
    CustomPlotlyChart,
    Scatter3DPlot,
)

from . import language
from .language.grammar import GrammarScore
from .language.openai_evals import OpenaiEval, PromptEval
from .language.embedding import Embedding
from .language.rouge import RougeScore
from .language.bleu import BLEUScore
from .language.meteor import METEORScore
from .language.text import (
    DocsLinkVersion,
    TextLength,
    TextComparison,
    KeywordDetector,
    WordCount,
)
from .language.model_grade import ModelGradeScore, OpenAIGradeScore
from .language.generation import (
    PromptGenerator,
    TextCompletion,
    OutputParser,
    TopicGenerator,
)
from .language.with_context import (
    ResponseFactualScore,
    ContextRelevance,
    ResponseCompleteness,
    ResponseRelevance,
    ResponseCompletenessWrtContext,
    ResponseConsistency,
    ResponseConciseness,
)
from .language.critique import LanguageCritique, ToneCritique
from .language.guideline import GuidelineAdherenceScore, PromptInjectionScore
from .language.conversation import ConversationSatisfactionScore
from .language.response_matching import ResponseMatchingScore, ValidResponseScore
from .language.topic import TopicAssignmentviaCluster

from . import io
from .io.base import CsvReader, JsonReader, DeltaReader, JsonWriter, DeltaWriter
from .io.excel import ExcelReader
from .io.bq import BigQueryReader, BigQueryWriter
from .io.mongodb import MongoDBReader
from .io.duck import DuckDBReader

from . import code
from .code.sql import (
    ParseCreateStatements,
    ParseSQL,
    ValidateTables,
    ExecuteAndCompareSQL,
)



import numpy as np

from typing import List, Union

from uptrain.v0.core.classes.distances import AbstractDistance


class HammingDistance(AbstractDistance):
    """Class that computes Hamming distance between base and reference vectors."""

    def check_valid_vector(self, vector) -> None:
        if not np.all(np.isin(vector, [0, 1])):
            raise Exception("Vector should contain only 0 and 1")

    def compute_distance(
        self,
        base: Union[List, np.ndarray],
        reference: Union[List, np.ndarray],
        normalized: bool = False,
    ) -> Union[int, np.ndarray]:
        """Compute the Hamming distance between base and reference vectors.
        Parameters
        ----------
        base
            It is the base vector for Hamming distance computation.
        reference
            It is the reference vector for Hamming distance computation.
        normalized : bool, optional
            If True, returns the normalized Hamming distance.
            If False, returns the regular Hamming distance.

        Returns
        -------
        int/np.ndarray
            The Hamming distance between base and reference.
        """

        base = np.array(base)
        reference = np.array(reference)

        self.check_valid_vector(base)
        self.check_valid_vector(reference)

        self.check_compatibility(base, reference)

        # Do not normalize if the base is a single bit
        if normalized and base.shape:
            return np.mean(base != reference, axis=tuple(range(1, len(base.shape))))

        return np.sum(base != reference, axis=tuple(range(1, len(base.shape))))



import sqlite3
from collections import defaultdict
from typing import Dict, Any, Set

import numpy as np
import polars as pl

from sqlglot import Expression
from sqlglot.expressions import Table, Column, ColumnDef, Create, Schema

from loguru import logger

PLACEHOLDER_TABLE = "PLACEHOLDER_TABLE"


def merge_dictionaries(dict1: Dict[Any, Set], dict2: Dict[Any, Set]):
    for key, value in dict2.items():
        if key in dict1:
            dict1[key] = dict1[key].union(value)
        else:
            dict1[key] = value

    return dict1


def extract_tables_and_columns(expression: Expression, alias_mapping=None):
    if alias_mapping is None:
        alias_mapping = {}

    tables = defaultdict(set)

    # Handle tables and table aliases
    if isinstance(expression, Table):
        table_key = f"{expression.db}.{expression.name}" if expression.db else expression.name
        if table_key not in tables:
            tables[table_key] = set([])
        if expression.alias:
            alias_mapping[expression.alias] = table_key

    for key, e in expression.iter_expressions():
        merge_dictionaries(tables, extract_tables_and_columns(e, alias_mapping))

    # Handle columns and their associated tables or table aliases
    if isinstance(expression, Column):
        table = f"{expression.db}.{expression.table}" if expression.db else expression.table
        if table and table in alias_mapping:
            table = alias_mapping[table]
        table = table or PLACEHOLDER_TABLE
        if table not in tables:
            tables[table] = set([])
        tables[table].add(expression.name)

    # Merge aliases
    # TODO: this does not work if table1 is alias of table2 which is an alias of table3
    for table in tables:
        if table in alias_mapping:
            true_table = alias_mapping[table]
            tables[true_table] = tables[true_table].union(tables[table])

    # remove alias tables
    for table in alias_mapping:
        if table in tables:
            del tables[table]
    return tables


def extract_tables_and_columns_from_create(expression: Create):
    schema: Schema = expression.this
    table = schema.this
    table_name = table.name
    columns = set([])
    for e in schema.expressions:
        if isinstance(e, ColumnDef):
            columns.add(e.name)
    return table_name, columns


def run_query(query, connection):
    cursor = connection.cursor()
    cursor.execute(query)

    # Get data and column names from the cursor
    data = cursor.fetchall()
    columns = [description[0] for description in cursor.description]

    df = pl.DataFrame(data, columns)
    return df


# Execute predicted SQL, ground truth and compute execution accuracy of the predicted sql. We assume these queries to be
# read queries.
def execute_and_compare_sql(predicted_sql, ground_truth, db_path, ignore_column_order=True, ignore_row_order=True):
    conn = sqlite3.connect(db_path)
    res = False
    try:
        # Run the queries
        pred_df = run_query(predicted_sql, conn)
        gt_df = run_query(ground_truth, conn)

        # Close the connection
        conn.close()

        if ignore_column_order:
            # Bring the columns to the same order
            pred_df = pred_df.select(sorted(pred_df.columns, key=str.lower))
            gt_df = gt_df.select(sorted(gt_df.columns, key=str.lower))

        if ignore_row_order:
            # Sort the dataframe rows to ignore row order
            pred_df = pred_df.sort(list(pred_df.columns))
            gt_df = gt_df.sort(list(gt_df.columns))

        res = np.array_equal(pred_df.to_numpy(), gt_df.to_numpy())
    except Exception as e:
        logger.warning(f"Error executing: {e}")
        pass
    return res




"""Operators to compute metrics over embeddings"""

from __future__ import annotations
import os
import uuid
from typing import Any, Callable, Literal, Union
from itertools import product
from functools import cache

import duckdb
import numba
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
from pydantic import BaseModel

from .ops_base import (
    Operator,
    OperatorExecutor,
    arrow_batch_to_table,
    array_arrow_to_np,
    array_np_to_arrow,
    np_arrays_to_arrow_table,
    table_arrow_to_np_arrays,
)

PARTITION_COLUMN = "partition_index"

# -----------------------------------------------------------
# Window Aggregation Operators (N-to-N)
# -----------------------------------------------------------


class WindowAggOp(BaseModel):
    """Operator to compute a window aggregation over a column in a table (N-to-N)."""

    value_col: str  # the column to aggregate over in each group
    seq_col: str  # the column to sort by before aggregation


@ray.remote
class WindowAggExecutor(OperatorExecutor):
    """Executor for window aggregations."""

    op: WindowAggOp
    cache_conn: duckdb.DuckDBPyConnection
    compute_loop: Callable

    def __init__(self, op: WindowAggOp, compute_op: Callable) -> None:
        """Initialize a Ray actor for the given operator.

        Args:
            op: Operator to execute.
            compute_op: Function to computes the aggregate value for the given value and intermediate value.
                The executor batches calls to this function and jit compiles the loop using numba.
        """
        self.op = op
        self.compute_loop = generate_compute_loop_for_window_agg(compute_op)

        os.makedirs(".cache", exist_ok=True)
        self.cache_conn = duckdb.connect(f".cache/{uuid.uuid4()}.duckdb")
        self.cache_conn.execute(
            f"CREATE TABLE interm_state (id STRING PRIMARY KEY, value FLOAT[]);"
        )  # TODO: support other types?

    def fetch_interm_state(self, tbl: pa.Table) -> pa.Table:
        """Fetches the intermediate state for the ids in the given table already seen by this operator,
        and selects the first value for each id not seen yet.
        """
        self.cache_conn.execute(
            """
            SELECT * FROM (
                SELECT id, value
                FROM interm_state
                WHERE id IN (SELECT {id_col} from tbl)
                
                UNION
            
                SELECT {id_col}, FIRST({value_col} ORDER BY {seq_col} ASC)
                FROM tbl 
                WHERE {id_col} NOT IN (SELECT id FROM interm_state)
                GROUP BY {id_col}
            )
            ORDER BY 1 ASC;
            """.format(
                id_col=PARTITION_COLUMN,
                value_col=self.op.value_col,
                seq_col=self.op.seq_col,
            )
        )
        return self.cache_conn.fetch_arrow_table()

    def update_interm_state(self, id_array: pa.Array, value_array: pa.Array):
        tbl = pa.Table.from_arrays([id_array, value_array], ["id", "value"])

        # upserts don't work for List data types, so we gotta do the update in steps.
        self.cache_conn.execute(
            "DELETE FROM interm_state WHERE id IN (SELECT id FROM tbl);"
        )
        self.cache_conn.execute("INSERT INTO interm_state SELECT id, value FROM tbl;")

    def run(self, tbl: Union[pa.Table, pa.RecordBatch]) -> pa.Array:
        tbl = arrow_batch_to_table(tbl)

        # We assume the table is already sorted by the partition_id and seq_col.
        # Get count of rows for each key value. The agg operator can use it to figure out the partition boundaries.
        partition_counts = (
            tbl.group_by([PARTITION_COLUMN])
            .aggregate([(PARTITION_COLUMN, "count")])
            .sort_by([(PARTITION_COLUMN, "ascending")])
        )
        partition_counts_array = array_arrow_to_np(
            partition_counts[f"{PARTITION_COLUMN}_count"]
        )

        [value_array] = table_arrow_to_np_arrays(tbl, [self.op.value_col])
        interm_state_table = self.fetch_interm_state(tbl)
        interm_state_array = table_arrow_to_np_arrays(interm_state_table, ["value"])[0]

        assert len(partition_counts_array) == len(interm_state_array)
        output_value_array, new_interm_state_array = self.compute_loop(
            partition_counts_array, value_array, interm_state_array
        )
        self.update_interm_state(
            interm_state_table["id"], array_np_to_arrow(new_interm_state_array)
        )
        return array_np_to_arrow(output_value_array)


def compute_op_l2dist_initial(
    value: np.ndarray, interm_value: np.ndarray
) -> tuple[np.floating, np.ndarray]:
    return np.sum(np.square(value - interm_value)), interm_value  # type: ignore


def compute_op_l2dist_running(
    value: np.ndarray, interm_value: np.ndarray
) -> tuple[np.floating, np.ndarray]:
    return np.sum(np.square(value - interm_value)), value  # type: ignore


def compute_op_cosine_dist_initial(
    value: np.ndarray, interm_value: np.ndarray
) -> tuple[np.floating, np.ndarray]:
    cos_similarity = np.dot(value, interm_value) / (
        np.linalg.norm(value) * np.linalg.norm(interm_value)
    )
    return (1 - cos_similarity), interm_value


def compute_op_cosine_dist_running(
    value: np.ndarray, interm_value: np.ndarray
) -> tuple[np.floating, np.ndarray]:
    cos_similarity = np.dot(value, interm_value) / (
        np.linalg.norm(value) * np.linalg.norm(interm_value)
    )
    return (1 - cos_similarity), value


def compute_op_norm_ratio_initial(
    value: np.ndarray, interm_value: np.ndarray
) -> tuple[np.floating, np.ndarray]:
    ratio = np.linalg.norm(value) / max(np.linalg.norm(interm_value), 1e-6)  # type: ignore
    return ratio, interm_value


def compute_op_norm_ratio_running(
    value: np.ndarray, interm_value: np.ndarray
) -> tuple[np.floating, np.ndarray]:
    ratio = np.linalg.norm(value) / max(np.linalg.norm(interm_value), 1e-6)  # type: ignore
    return ratio, value


@cache
def generate_compute_loop_for_window_agg(compute_fn: Callable) -> Callable:
    """Generates a compute loop for the given function."""

    inlined_fn = numba.njit(inline="always")(compute_fn)

    @numba.njit
    def compute_loop(partition_counts_array, value_array, interm_state_array):
        """Goes through the value array and computes the output value for each value, and updating the
        intermediate state.

        Args:
            partition_counts_array (np.ndarray): Array of counts of rows for each partition.
            value_array (np.ndarray): Array of values to compute the output for.
            interm_state_array (np.ndarray): Array of intermediate states for each partition.

        Returns:
            tuple[np.ndarray, np.ndarray]: The output value array and the new intermediate state array.
        """
        output_value_array = np.zeros(len(value_array))
        new_interm_state_array = np.zeros_like(interm_state_array)

        interm_idx = 0
        interm_state = interm_state_array[interm_idx]
        change_indices = np.cumsum(partition_counts_array)

        for idx, value in enumerate(value_array):
            if idx == change_indices[interm_idx]:
                # store interm state for the current partition, and move to the next
                new_interm_state_array[interm_idx] = interm_state
                interm_idx += 1
                interm_state = interm_state_array[interm_idx]
            output_value_array[idx], interm_state = inlined_fn(value, interm_state)  # type: ignore
        new_interm_state_array[interm_idx] = interm_state

        return output_value_array, new_interm_state_array

    return compute_loop


class L2Dist(WindowAggOp):
    """Computes the L2 distance between the current embedding and the initial/previous one."""

    mode: Literal["initial", "running"]

    def make_actor(self):
        # string casting to appease pylance since it doesn't let me compare a Literal to a string
        mode = str(self.mode)
        if mode == "initial":
            return WindowAggExecutor.remote(self, compute_op_l2dist_initial)
        elif mode == "running":
            return WindowAggExecutor.remote(self, compute_op_l2dist_running)
        else:
            raise ValueError(f"Unknown mode: {mode}")


class CosineDist(WindowAggOp):
    """Computes the cosine distance between the current embedding and the initial/previous one."""

    mode: Literal["initial", "running"]

    def make_actor(self):
        mode = str(self.mode)
        if mode == "initial":
            return WindowAggExecutor.remote(self, compute_op_cosine_dist_initial)
        elif mode == "running":
            return WindowAggExecutor.remote(self, compute_op_cosine_dist_running)
        else:
            raise ValueError(f"Unknown mode: {mode}")


class NormRatio(WindowAggOp):
    """Computes the ratio between the current embedding and the initial/previous one."""

    mode: Literal["initial", "running"]

    def make_actor(self):
        mode = str(self.mode)
        if mode == "initial":
            return WindowAggExecutor.remote(self, compute_op_norm_ratio_initial)
        elif mode == "running":
            return WindowAggExecutor.remote(self, compute_op_norm_ratio_running)
        else:
            raise ValueError(f"Unknown mode: {mode}")


# -----------------------------------------------------------
# Pairwise metrics computed over the full table at once (N-to-1)
# -----------------------------------------------------------


class ReduceOp(BaseModel):
    """Operators that produce a single row of output (N-to-1)."""

    value_col: str  # the column to aggregate over


class CosineDistHistogram(ReduceOp):
    """Computes the histogram of cosine distances between all embeddings in the table."""

    def make_actor(self):
        return PairwiseMetricExecutor.remote(self, compute_op_cosine_dist_initial)


@ray.remote
class PairwiseMetricExecutor:
    """Executor for computing a metric between pairs of embeddings."""

    op: ReduceOp
    compute_func: Callable

    def __init__(self, op: ReduceOp, compute_op: Callable) -> None:
        self.op = op
        self.compute_fn = compute_op


# -----------------------------------------------------------
# Partition Operator
# -----------------------------------------------------------


class PartitionOp(BaseModel, arbitrary_types_allowed=True):
    """Operator to partition a table/batch based on values in a set of columns,
    compute an aggregate on each partition, and combine the results.
    """

    columns: list[str]  # list of columns to partition on
    agg_ops: dict[
        str, Operator
    ]  # operators to apply to each partition, they must have the same n-arity

    def make_actor(self):
        return PartitionExecutor.remote(self)


@ray.remote
class PartitionExecutor:
    """Executor for partitioning."""

    op: PartitionOp
    agg_actors: list[OperatorExecutor]

    def __init__(self, op: PartitionOp) -> None:
        self.op = op
        self.agg_actors = [op.make_actor() for op in op.agg_ops.values()]

    def run(self, batch: Union[pa.Table, pa.RecordBatch]) -> pa.Table:
        if batch is None:
            return None
        tbl = arrow_batch_to_table(batch)

        # we don't know the primary key, so all aggregations must have the same sequence column so we can join back the outputs
        seq_cols = [op.seq_col for op in self.op.agg_ops.values()]  # type: ignore
        assert len(set(seq_cols)) == 1
        seq_col = seq_cols[0]

        # Create a column by concatenating the partition columns, and use it as the sort key so partitions are contiguous
        tbl = tbl.append_column(
            PARTITION_COLUMN,
            pc.binary_join_element_wise(  # type: ignore
                *[pc.cast(tbl[col], pa.string()) for col in self.op.columns], "_"
            ),
        ).sort_by([(PARTITION_COLUMN, "ascending"), (seq_col, "ascending")])

        result_refs = []
        for agg_op, agg_actor in zip(self.op.agg_ops.values(), self.agg_actors):
            assert isinstance(agg_op, WindowAggOp)
            tbl_input = tbl.select([PARTITION_COLUMN, agg_op.seq_col, agg_op.value_col])
            result_refs.append(agg_actor.run.remote(tbl_input))

        output_arrays = [arr for arr in ray.get(result_refs)]
        for op, arr in zip(self.op.agg_ops, output_arrays):
            tbl = tbl.append_column(op, arr)  # type: ignore
        return tbl



import numpy as np

from river import drift

from uptrain.v0.core.classes.monitors import AbstractMonitor
from uptrain.v0.constants import DataDriftAlgo, MeasurableType
from uptrain.v0.core.classes.measurables import MeasurableResolver
from uptrain.v0.constants import Monitor
from uptrain.v0.core.lib.helper_funcs import add_data_to_warehouse


class ConceptDrift(AbstractMonitor):
    monitor_type = Monitor.CONCEPT_DRIFT

    def base_init(self, fw, check):
        if check.get("measurable_args", None):
            self.measurable = MeasurableResolver(check["measurable_args"]).resolve(fw)
        else:
            self.measurable = MeasurableResolver(
                {"type": MeasurableType.ACCURACY}
            ).resolve(fw)

        self.avg_acc = 0
        self.drift_alerted = False
        self.algorithm = check["algorithm"]
        self.counter = 0
        self.plot_name = f"avg_accuracy_{self.measurable.col_name()}"
        self.feat_slicing = check.get("feat_slicing", False)

        if self.algorithm == DataDriftAlgo.DDM:
            warm_start = check.get("warm_start", 500)
            warn_threshold = check.get("warn_threshold", 2.0)
            alarm_threshold = check.get("alarm_threshold", 3.0)
            self.algo = drift.DDM(warm_start, warn_threshold, alarm_threshold)
        elif self.algorithm == DataDriftAlgo.ADWIN:
            delta = check.get("delta", 0.002)
            clock = check.get("clock", 32)
            max_buckets = check.get("max_buckets", 5)
            min_window_length = check.get("min_window_length", 5)
            grace_period = check.get("grace_period", 5)
            self.algo = drift.ADWIN(
                delta, clock, max_buckets, min_window_length, grace_period
            )
        else:
            raise Exception("Data drift algo type not supported")

    def need_ground_truth(self):
        return True

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        all_models = [
            x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
            for x in self.model_measurables
        ]
        all_features = [
            x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
            for x in self.feature_measurables
        ]

        batch_acc = self.measurable.compute_and_log(inputs, outputs, gts, extra_args)
        batch_acc = self._preprocess(batch_acc)

        for i, acc in enumerate(batch_acc):
            alert = None
            self.algo.update(acc)

            if self.algo.drift_detected and not self.drift_alerted:
                alert = f"Concept Drift detected! [Algorithm: {self.algorithm}] [Time: {self.counter}] [Measurable: {self.measurable.col_name()}]"
                print(alert)
                self.drift_alerted = True

            self.avg_acc = (self.avg_acc * self.counter + acc) / (self.counter + 1)
            self.counter += 1
            models = dict(
                zip(
                    self.model_names,
                    [all_models[jdx][i] for jdx in range(len(self.model_names))],
                )
            )
            features = dict(
                zip(
                    self.feature_names,
                    [all_features[jdx][i] for jdx in range(len(self.feature_names))],
                )
            )
            self.log_handler.add_scalars(
                self.plot_name,
                {"y_avg_accuracy": self.avg_acc},
                self.counter,
                self.dashboard_name,
                features=features,
                models=models,
            )

            if isinstance(alert, str):
                self.log_handler.add_alert(
                    "Model Performance Degradation Alert 🚨", alert, self.dashboard_name
                )
        if self.feat_slicing:
            add_data_to_warehouse(
                {"id": inputs["id"], self.dashboard_name: batch_acc},
                self.path_dashboard_data,
            )

    def _preprocess(self, batch):
        if self.algorithm == DataDriftAlgo.DDM:
            return np.array([0 if x else 1 for x in batch])
        else:
            return np.array(batch)



from abc import ABC
from typing import Any


class AbstractMeasurable(ABC):
    """Abstract base class for objects that can be measured or evaluated.

    This abstract class provides a template for defining objects that can
    be evaluated in some way. Subclasses of  AbstractMeasurable should
    implement the `_compute` method, which performs the actual computation.
    """

    def __init__(self) -> None:
        super().__init__()

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> Any:
        """Computes the measurement or evaluation of this object.

        This method should be overridden by subclasses to perform the actual
        computation of the measurement or evaluation. The inputs, outputs and
        gts parameters can be used in the computation. The extra parameter can
        be used to provided additional information for the computation.

        Parameters
        ----------
        inputs
            Inputs values to use in the computation
        outputs
            Outputs values to use in the computation
        gts
            Ground truth values to use in the computation
        extra
            Additional information to use in the computation

        Returns
        -------
        Any
            Result of the computation
        """
        raise Exception("Should be defined for each individual class")

    def col_name (self) -> str:
        """Returns the column name of the Measurable"""
        raise Exception("Should be defined for each individual class")



from __future__ import annotations
import csv, _csv
import os
import shutil
import json
from typing import IO, TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from uptrain.v0.core.classes.framework import Framework
    from uptrain.v0.core.classes.helpers.config_handler import Config

from uptrain.v0.core.lib.helper_funcs import make_dir_friendly_name, clear_directory
from uptrain.v0.core.encoders import NumpyEncoder
from uptrain.v0.core.classes.logging.new_st_setup import StreamlitRunner


# -----------------------------------------------------------
# Implement LogWriters for different file formats
# -----------------------------------------------------------
def get_logs_addr_for_check(
    log_folder: str, dashboard_name: str, distance_type: str, reference: str
) -> str:
    """get the path to the log file for the dashboard."""
    dashboard_name = make_dir_friendly_name(dashboard_name)
    plot_name = make_dir_friendly_name(distance_type + "_" + str(reference))
    return os.path.join(log_folder, dashboard_name, f"{plot_name}.csv")


class LogWriter(Protocol):
    def log(self, data: dict):
        """Write data to a log file. Expects to get called multiple times."""
        ...


class CsvWriter(LogWriter):
    """
    NOTE: does this definitely work for multiple classes with a handle to the same CSV file?
    # RESPONSE: No, it didn't. Flushing after each write still leaves blank lines in the CSV.
    # SOLVED using memoization on the `make_logger` method, one file for each set of unique args to it.
    """

    file_handle: IO[str]
    writer: "_csv._writer"
    headers: list[str]
    initialized: bool

    def __init__(self, fname: str) -> None:
        self.file_handle = open(fname, "a")
        self.writer = csv.writer(self.file_handle)
        self.initialized = False
        self.headers = []

    def __del__(self):
        self.file_handle.close()

    def log(self, data: dict):
        """Create a csv file and write the header to it if it doesn't exist.
        Post that, write the data to the file.
        """
        if not self.initialized:
            self.initialized = True
            self.headers = list(data.keys())
            self.writer.writerow(self.headers)
        self.writer.writerow([data[k] for k in self.headers])


class JsonWriter(LogWriter):
    fname: str

    def __init__(self, fname: str) -> None:
        self.fname = fname

    def log(self, data: dict):
        """create a json file and write the data to it. Or append if it already exists."""
        with open(self.fname, "a") as f:
            json.dump(data, f, cls=NumpyEncoder)
            f.write("\n")


# -----------------------------------------------------------
# LogHandler object
# -----------------------------------------------------------


class LogHandler:
    list_writers: dict[str, "LogWriter"]  # cache writer objects for each file

    def __init__(self, framework: "Framework", cfg: "Config"):
        self.fw = framework
        self.path_all_data = framework.path_all_data

        self.log_folder = cfg.logging_args.log_folder
        if os.path.exists(self.log_folder):
            print(f"Deleting contents of the log folder at: {self.log_folder}")
            clear_directory(self.log_folder)
        else:
            print(f"Creating the log folder at: {self.log_folder}")
            os.makedirs(self.log_folder, exist_ok=True)
        self.list_writers = dict()

        # serialize the config to a json file so the consumers can read it
        cfg_file = os.path.join(self.log_folder, "config.json")
        with open(cfg_file, "w") as f:
            f.write(cfg.json())

        # initialize the streamlit dashboard
        self.st_runner = StreamlitRunner(self.log_folder, cfg.logging_args.dashboard_port)
        if cfg.logging_args.run_background_streamlit:
            self.st_runner.start()
        else:
            print(
                "To start the streamlit dashboard, run the following command: ",
                self.st_runner.launch_cmd,
            )

        # Get Webhook URL for alerting on slack
        self.webhook_url = cfg.logging_args.slack_webhook_url

    def make_logger(
        self, dashboard_name: str, plot_name: str, fmt: str = "csv"
    ) -> LogWriter:
        """Initialize a CSV logger that the `check` object can write data to."""
        dashboard_name = make_dir_friendly_name(dashboard_name)
        plot_name = make_dir_friendly_name(plot_name)

        dir_name = os.path.join(self.log_folder, dashboard_name)
        os.makedirs(dir_name, exist_ok=True)
        fname = os.path.join(dir_name, f"{plot_name}.{fmt}")

        if fname not in self.list_writers:
            if fmt == "json":
                self.list_writers[fname] = JsonWriter(fname)
            elif fmt == "csv":
                self.list_writers[fname] = CsvWriter(fname)
            else:
                raise ValueError(f"Unknown format for the log-file: {fmt}")
        return self.list_writers[fname]

    def add_alert(self, alert_name, alert, dashboard_name):
        dashboard_name = make_dir_friendly_name(dashboard_name)
        dashboard_dir = os.path.join(self.log_folder, dashboard_name)
        plot_folder = os.path.join(dashboard_dir, "alerts")
        os.makedirs(plot_folder, exist_ok=True)

        file_name = os.path.join(plot_folder, str(alert_name) + ".json")
        with open(file_name, "w") as f:
            json.dump(alert, f)

        if self.webhook_url:
            message = (
                f"Dashboard: {dashboard_name}, Alert name: {alert_name}, Alert: {alert}"
            )
            self.slack_notification({"text": message})

    def slack_notification(self, message):
        import urllib3

        try:
            http = urllib3.PoolManager()
            response = http.request(
                "POST",
                self.webhook_url,
                body=json.dumps(message),
                headers={"Content-Type": "application/json"},
                retries=False,
            )
        except Exception as e:
            print("Caught Exception")
            print(e)



import numpy as np

from typing import List, Union

from uptrain.v0.core.classes.distances import AbstractDistance


class NormRatio(AbstractDistance):
    """Class that computes the normalized ratio between base and reference vectors."""

    def __init__(self, norm_min: float = 1e-6):
        self.norm_min = norm_min

    def calculate_norm(self, vector) -> float:
        if len(vector.shape) > 1:
            return np.linalg.norm(vector, axis=1)
        return np.abs(vector)

    def compute_distance(
        self, base: Union[List, np.ndarray], reference: Union[List, np.ndarray]
    ) -> np.ndarray:
        """Compute the norm ratio between base and reference vectors.

        Parameters
        ----------
        base
            It is the base vector for norm ratio computation.
        reference
            It is the reference vector for norm ratio computation.

        Returns
        -------
        np.ndarray
            The norm ratio between base and reference.
        """

        base = np.array(base)
        reference = np.array(reference)

        self.check_compatibility(base, reference)

        base_norm = self.calculate_norm(base)
        ref_norm = self.calculate_norm(reference)

        return base_norm / np.maximum(ref_norm, self.norm_min)



import matplotlib.pyplot as plt
import numpy as np
import uptrain.v0 as v0

from scipy.stats import zscore


def plot_graph(exam_scores):
    """Plot z-score graph."""

    z_scores = zscore(exam_scores)
    plt.hist(z_scores, bins=50, alpha=0.5, label="Z score", color="green")

    # Add a legend and labels
    plt.legend()
    plt.xlabel("Z score")
    plt.ylabel("Frequency")
    plt.title("Distribution of Z scores with outliers")

    # Show the plot
    plt.show()


def test_data_integrity_zscore():
    """Test data integrity z-score.

    The test generates random exam scores, adds some random outliers to the data,
    and shuffles the data. Then, the UpTrain framework is used to monitor the
    data integrity of the generated data using z-score.

    To generate the test, 5000 random exam scores are generated with a mean of
    0 and a standard deviation of 250. Then, 250 random outliers are added
    to the data. The outliers account for about 2-5%  of the total data depending
    on the randomly generated values.

    The test passes if the monitored z-score data integrity detects the presence
    of outliers in the data.
    """

    random_state = np.random.RandomState(seed=1337)
    mean = 0
    std = 250
    num_samples = 5000
    exam_scores = random_state.normal(mean, std, num_samples)

    # Add some random outliers that are not normally distributed
    num_outliers = 250
    outliers = np.concatenate(
        (
            random_state.uniform(-1500, -500, num_outliers // 2),
            random_state.uniform(500, 1500, num_outliers // 2),
        )
    )

    # Add the outliers to the exam scores
    exam_scores = np.concatenate((exam_scores, outliers))

    # Shuffle the scores
    # random_state.shuffle(exam_scores)

    # plot_graph(exam_scores)

    # Create a configuration for the framework
    cfg = {
        # Define checks to be performed
        "checks": [
            {
                "type": v0.Monitor.DATA_INTEGRITY,
                "measurable_args": {
                    "type": v0.MeasurableType.INPUT_FEATURE,
                    "feature_name": "scores",
                },
                "integrity_type": "z_score",
                "threshold": 3,
            }
        ],
        # Specify where the data from z-score data integrity should be stored
        "retraining_folder": "uptrain_smart_data_integrity_zscore",
        # # Specify logging arguments
        # "st_logging" should be True if we want streamlit logging, False otherwise
        "logging_args": {
            "st_logging": True,
            "log_folder": "uptrain_data_integrity_zscore",
        },
    }

    framework = v0.Framework(cfg)
    batch_size = 64
    size = len(exam_scores)

    # Feed the data to the framework
    for i in range(size // batch_size):
        framework.log(
            inputs={
                "scores": exam_scores[i * batch_size : (i + 1) * batch_size],
            }
        )



import os
import sys
import polars as pl

from uptrain.utilities import lazy_load_dep
st = lazy_load_dep("streamlit", "streamlit>=1.23")

from uptrain.operators import ColumnOp
from uptrain.framework import Settings, CheckSet, Check
from uptrain.dashboard.st_helpers import (
    st_make_check_selector,
    st_show_table,
    st_setup_layout,
    read_checkset,
    read_settings,
    CONSOLIDATED_CHECK,
)

# -----------------------------------------------------------
# Utility functions
# -----------------------------------------------------------

# TODO: Streamlit can't pickle custom operators, so we need to use a workaround to cache
# read_checkset = st.cache_data(read_checkset)


@st.cache_data
def load_data_for_check_local(_settings: "Settings", _check: "Check", check_name: str):
    from uptrain.operators import DeltaWriter, JsonWriter

    sink = CheckSet._get_sink_for_check(_settings, _check)
    if isinstance(sink, (DeltaWriter, JsonWriter)):
        source = sink.to_reader()
        source.setup(_settings)
    else:
        raise NotImplementedError(f"{type(sink)} is not supported for now.")
    output = source.run()
    return output["output"]


@st.cache_data
def load_data_for_checks_consolidated(_settings: "Settings", _checkset: "CheckSet"):
    all_cols = {}
    for _check in _checkset.checks:
        if not all(isinstance(op, ColumnOp) for op in _check.operators):
            continue
        data = load_data_for_check_local(_settings, _check, _check.name)
        for col in data.columns:
            all_cols[col] = data[col]

    return pl.DataFrame(all_cols)


# -----------------------------------------------------------
# Run code
# -----------------------------------------------------------


# Set up layout of the dashboard and the sidebar
st_setup_layout()

# Read parameters for this run
LOGS_DIR = sys.argv[1]
CHECK_SET = read_checkset(LOGS_DIR)
SETTINGS = read_settings(LOGS_DIR)

# Pick a check to view
check = st_make_check_selector(CHECK_SET)

# Check the plot operator for the check
plot_ops = check.plots
if len(plot_ops) == 0:
    st.warning("No plot operator specified for the check")
    st.stop()

# load data
if check.name == CONSOLIDATED_CHECK.name:
    data = load_data_for_checks_consolidated(SETTINGS, CHECK_SET)
else:
    data = load_data_for_check_local(SETTINGS, check, check.name)
if data is None:
    st.error("No data found per the specified config.")
    st.stop()

# Plot the data
for plot_op in plot_ops:
    st.header(plot_op.title)
    if plot_op.kind == "table":
        st_show_table(data)
    else:
        plot_op.setup(SETTINGS)
        output = plot_op.run(data)["extra"]["chart"]
        st.plotly_chart(output)
    st.divider()
